//+--------------------------------------------------------------------------
//|
//| rcreader.cpp - a parser for command line arguments
//|
//| This file is part of a larger package; see the README supplied with
//| this for information about the package, how to build and how to run
//| it.
//|
//| (c) 2002, 2020 Dietmar Koenig <http://www.faoileag.de>
//|
//| This program is free software: you can redistribute it and/or modify
//| it under the terms of the GNU General Public License as published by
//| the Free Software Foundation, either version 3 of the License, or
//| (at your option) any later version.
//|
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//| GNU General Public License for more details.
//|
//| You should have received a copy of the GNU General Public License
//| along with this program.  If not, see <https://www.gnu.org/licenses/>.
//|
//+--------------------------------------------------------------------------
#include "rcreader.h"

//+--------------------------------------------------------------------------+
//|                                                                          |
//| args_t parseargs()                                                       |
//|                                                                          |
//| parses all command line arguments and fills the respective fields in the |
//| structure.                                                               |
//| the memory allocated for the struct must be freed in the main() function |
//| calling parseargs().                                                     |
//| returns a args_t structure with the values taken from the command line   |
//| arguments.                                                               |
//|                                                                          |
//+--------------------------------------------------------------------------+
args_t* parseargs(int argc, char** argv) {
	args_t* args = new args_t();
	// set the fields on default values in case that argument is not used:
	args->immediate = XMANDELBROTARGS_DEFAULTIMMEDIATE;
	args->verbosity    = XMANDELBROTARGS_DEFAULTVERBOSITY;
	args->infinity     = XMANDELBROTARGS_DEFAULTINFINITY;
	args->x            = XMANDELBROTARGS_DEFAULTX;
	args->y            = XMANDELBROTARGS_DEFAULTY;
	args->width        = XMANDELBROTARGS_DEFAULTWIDTH;
	args->height       = XMANDELBROTARGS_DEFAULTHEIGHT;
	args->canvaswidth  = 0; // if zero, -canvas arg is not used
	args->canvasheight = 0; // "  "     "       "   "  "   "
	args->noaspect     = XMANDELBROTARGS_DEFAULTNOASPECT;
	args->function     = XMANDELBROTARGS_DEFAULTFUNCTION;
	args->pmin         = XMANDELBROTARGS_DEFAULTPMIN;
	args->pmax         = XMANDELBROTARGS_DEFAULTPMAX;
	args->qmin         = XMANDELBROTARGS_DEFAULTQMIN;
	args->qmax         = XMANDELBROTARGS_DEFAULTQMAX;
	strnzcpy(&args->title[0], XMANDELBROTARGS_DEFAULTTITLE, sizeof(args->title));
	memset(&args->progname[0], 0, sizeof(args->progname));
	memset(&args->filename[0], 0, sizeof(args->filename));

	// parse the command line arguments
	strnzcpy(args->progname, argv[0], sizeof(args->progname));
	while(argc > 1) {
		if ((strcmp(argv[1], "-c") == 0) ||
						 (strcmp(argv[1], "-canvas") == 0)) {
			if ((argc == 2) || (argv[2][0] == '-')){
				fprintf(stderr, "%s: argument missing.\n", argv[1]);
				exit(-1);
    	}
			// check for invalid characters:
			for (unsigned int i=0; i<strlen(argv[2]); ++i) {
				if ((!isdigit(argv[2][i])) &&
						(argv[2][i] != 'x')) {
      		fprintf(stderr, "%s: argument invalid.\n", argv[1]);
					exit(-1);
				}
			}
			// get the canvas width:
			char* str = strtok(argv[2], "x");
			if (str == NULL) {
     		fprintf(stderr, "%s: argument invalid.\n", argv[1]);
				exit(-1);
			}
			args->canvaswidth = atoi(str);
			if (args->canvaswidth > XMANDELBROTARGS_MAXWIDTH) {
				fprintf(stderr, "%s: argument out of range.\n", argv[1]);
				exit(-1);
			}
			// get the canvas height:
			str = strtok(NULL, "x");
			if (str == NULL) {
     		fprintf(stderr, "%s: argument invalid.\n", argv[1]);
				exit(-1);
			}
			args->canvasheight = atoi(str);
			if (args->canvasheight > XMANDELBROTARGS_MAXHEIGHT) {
				fprintf(stderr, "%s: argument out of range.\n", argv[1]);
				exit(-1);
			}
			--argc;
			++argv;
		}
		else if ((strcmp(argv[1], "-d") == 0) ||
							(strcmp(argv[1], "-debug") == 0)) {
			args->verbosity = 2;
		}
		else if ((strcmp(argv[1], "-f") == 0) ||
						 (strcmp(argv[1], "-file") == 0)) {
			if ((argc == 2) || (argv[2][0] == '-')){
				fprintf(stderr, "%s: argument missing.\n", argv[1]);
				exit(-1);
    	}
			if (!strcmp(argv[2], "")) {
				fprintf(stderr, "Joking, eh?\n");
				exit(-1);
			}
			strnzcpy(args->filename, argv[2], sizeof(args->filename));
    	--argc;
			++argv;
		}
		else if (strcmp(argv[1], "-geometry") == 0) {
			if ((argc == 2) || (argv[2][0] == '-')){
				fprintf(stderr, "%s: argument missing.\n", argv[1]);
				exit(-1);
    	}
			// check for invalid characters:
			for (unsigned int i=0; i<strlen(argv[2]); ++i) {
				if ((!isdigit(argv[2][i])) &&
						(argv[2][i] != 'x') &&
						(argv[2][i] != '+')) {
      		fprintf(stderr, "%s: argument invalid.\n", argv[1]);
					exit(-1);
				}
			}
			// get the width:
			char* str = strtok(argv[2], "x");
			if (str == NULL) {
     		fprintf(stderr, "%s: argument invalid.\n", argv[1]);
				exit(-1);
			}
			args->width = atoi(str);
			if (args->width > XMANDELBROTARGS_MAXWIDTH) {
				fprintf(stderr, "%s: argument out of range.\n", argv[1]);
				exit(-1);
			}
			// get the height:
			str = strtok(NULL, "+");
			if (str == NULL) {
     		fprintf(stderr, "%s: argument invalid.\n", argv[1]);
				exit(-1);
			}
			args->height = atoi(str);
			if (args->height > XMANDELBROTARGS_MAXHEIGHT) {
				fprintf(stderr, "%s: argument out of range.\n", argv[1]);
				exit(-1);
			}
			// get distance from left:
			str = strtok(NULL, "+");
			if (str == NULL) {
     		fprintf(stderr, "%s: argument invalid.\n", argv[1]);
				exit(-1);
			}
			args->x = atoi(str);
			if (args->x > (XMANDELBROTARGS_MAXWIDTH-args->width)) {
				fprintf(stderr, "%s: argument out of range.\n", argv[1]);
				exit(-1);
			}
			// get distance from top:
			str = strtok(NULL, "+");
			if (str == NULL) {
     		fprintf(stderr, "%s: argument invalid.\n", argv[1]);
				exit(-1);
			}
			args->y = atoi(str);
			if (args->y > (XMANDELBROTARGS_MAXHEIGHT-args->height)) {
				fprintf(stderr, "%s: argument out of range.\n", argv[1]);
				exit(-1);
			}
			--argc;
			++argv;
		} // end else if (strcmp(argv[1], "-geometry") == 0)
		else if ((strcmp(argv[1], "-h") == 0) ||
						 (strcmp(argv[1], "-help") == 0)) {
			fprintf(stdout, "usage: \%s [options...]\n\n", args->progname);
			fprintf(stdout, "\twith the following options being possible:\n\n");
			fprintf(stdout, "\t-c, -canvas <WxH>     tells the program to paint on a canvas\n");
			fprintf(stdout, "\t                      of this size; if the given size is larger\n");
			fprintf(stdout, "\t                      than the window size, the excess is just\n");
			fprintf(stdout, "\t                      \"painted\" into the video memory, if it is\n");
			fprintf(stdout, "\t                      smaller, the canvas is centered in the window.\n");			
			fprintf(stdout, "\t-d, -debug            tells the program to be _very_ verbose.\n");
			fprintf(stdout, "\t-f, -file <filename>  specifies file for loading / saving of\n");
			fprintf(stdout, "\t                      snapshot data.\n");
			fprintf(stdout, "\t-geometry <WxH+X+Y>   sets the window's size and position.\n");
			fprintf(stdout, "\t                      W: width in pixel (W < %d).\n", XMANDELBROTARGS_MAXWIDTH);
			fprintf(stdout, "\t                      H: height in pixel (H < %d).\n", XMANDELBROTARGS_MAXHEIGHT);
			fprintf(stdout, "\t                      X: upper left corner x (X < %d-W)\n", XMANDELBROTARGS_MAXWIDTH);
			fprintf(stdout, "\t                      Y: upper left corner y (Y < %d-H)\n", XMANDELBROTARGS_MAXHEIGHT);
			fprintf(stdout, "\t-h, -help             prints this text.\n");
			fprintf(stdout, "\t-i, -immediate        tells %s to start drawing immediately.\n", args->progname);
			fprintf(stdout, "\t-infinity <int>       sets the infinity threshold (%d < int < %d).\n", XMANDELBROTARGS_MININFINITY, XMANDELBROTARGS_MAXINFINITY);
			fprintf(stdout, "\t-julia                tells the program to draw a julia set.\n");
			fprintf(stdout, "\t-mandelbrot           tells the program to draw a mandelbrot set.\n");
			fprintf(stdout, "\t-noaspect             tells %s not to maintain the aspect\n", args->progname);
			fprintf(stdout, "\t                      ratio given by the canvas (see -c) when zooming\n");
			fprintf(stdout, "\t                      into the set by selecting a rectangular area.\n");
			fprintf(stdout, "\t                      this can lead to distorted images.\n");
			fprintf(stdout, "\t-pmin <+-x.xxxx>      sets the left boundary (%1.2f < pmin < pmax < %1.2f).\n", XMANDELBROTARGS_MINP, XMANDELBROTARGS_MAXP);
			fprintf(stdout, "\t-pmax <+-x.xxxx>      sets the right boundary.\n");
			fprintf(stdout, "\t-qmax <+-y.yyyy>      sets the upper boundary (%1.2f < qmin < qmax < %1.2f).\n", XMANDELBROTARGS_MINQ, XMANDELBROTARGS_MAXQ);
			fprintf(stdout, "\t-qmin <+-y.yyyy>      sets the lower boundary.\n");
			fprintf(stdout, "\t-t, -title <title>    sets the window's title.\n");
			fprintf(stdout, "\t-v, -verbose          tells the program to print some runtime information to\n");
			fprintf(stdout, "\t                      stdout.\n");
			fprintf(stdout, "\n\n");
			fprintf(stdout, "\tThe following keys can be used at runtime:\n\n");
			fprintf(stdout, "\t[Escape]              Clear - clears a zoom in rectangle.\n");
			fprintf(stdout, "\t[d]                   Draw  - redraws the current image.\n");
			fprintf(stdout, "\t[l]                   Load  - loads snapshot data and draws the\n");
			fprintf(stdout, "\t                              image generated by this data.\n");
			fprintf(stdout, "\t[p]                   Print - prints the current data to stdout.\n");
			fprintf(stdout, "\t[q]                   Quit  - ends the program.\n");
			fprintf(stdout, "\t[s]                   Save  - saves snapshot data.\n");
			fprintf(stdout, "\n\n");
			fprintf(stdout, "\tZoom in (draw a new image):\n\n");
			fprintf(stdout, "\tTo zoom in, press the left mouse button when the mouse pointer\n");
			fprintf(stdout, "\tis over the desired top left corner, keep it pressed while moving\n");
			fprintf(stdout, "\tthe mouse to create a rectangle of the desired size.\n");
			fprintf(stdout, "\tWhen finished, release left mouse button. Now you can tell %s\n", args->progname);
			fprintf(stdout, "\tto draw a new image with the borders specified by the rectangle by\n");
			fprintf(stdout, "\teither clicking the right mouse button or by pressing the key [d].\n");
			fprintf(stdout, "\tThe [Escape] key resets, i.e. the selected rectangle is cleared is\n");
			fprintf(stdout, "\tnot used as the base for the next image.\n");
			exit(0);
  	}
		else if ((strcmp(argv[1], "-i") == 0) ||
							(strcmp(argv[1], "-immediate") == 0)) {
			args->immediate = 1;
		}
		else if (strcmp(argv[1], "-infinity") == 0) {
			if ((argc == 2) || (argv[2][0] == '-')){
				fprintf(stderr, "%s: argument missing.\n", argv[1]);
				exit(-1);
    	}
			args->infinity = atoi(argv[2]);
			if ((args->infinity < XMANDELBROTARGS_MININFINITY) ||
					(args->infinity > XMANDELBROTARGS_MAXINFINITY)) {
				fprintf(stderr, "%s: argument out of range.\n", argv[1]);
				exit(-1);
			}
    	--argc;
			++argv;
		}
		else if (strcmp(argv[1], "-julia") == 0) {
			args->function = XMANDELBROTARGS_FUNCTIONJULIA;
		}
		else if (strcmp(argv[1], "-mandelbrot") == 0) {
			args->function = XMANDELBROTARGS_FUNCTIONJULIA;
		}
		else if (strcmp(argv[1], "-noaspect") == 0) {
			args->noaspect = 1;
		}
		else if (strcmp(argv[1], "-pmin") == 0) {
			if ((argc == 2) || ((argv[2][0] == '-') && (!isdigit(argv[2][1])))){
				fprintf(stderr, "%s: argument missing.\n", argv[1]);
				exit(-1);
    	}
			// check for invalid characters:
			for (unsigned int i=0; i<strlen(argv[2]); ++i) {
				if ((!isdigit(argv[2][i])) &&
						(argv[2][i] != '.') &&
						(argv[2][i] != '+') &&
						(argv[2][i] != '-')) {
      		fprintf(stderr, "%s: argument invalid.\n", argv[1]);
					exit(-1);
				}
			}
			args->pmin = atof(argv[2]);
			if ((args->pmin < XMANDELBROTARGS_MINP) ||
					(args->pmin > XMANDELBROTARGS_MAXP)) {
				fprintf(stderr, "%s: argument out of range.\n", argv[1]);
				exit(-1);
			}
    	--argc;
			++argv;
		}
		else if (strcmp(argv[1], "-pmax") == 0) {
			if ((argc == 2) || ((argv[2][0] == '-') && (!isdigit(argv[2][1])))){
				fprintf(stderr, "%s: argument missing.\n", argv[1]);
				exit(-1);
    	}
			// check for invalid characters:
			for (unsigned int i=0; i<strlen(argv[2]); ++i) {
				if ((!isdigit(argv[2][i])) &&
						(argv[2][i] != '.') &&
						(argv[2][i] != '+') &&
						(argv[2][i] != '-')) {
      		fprintf(stderr, "%s: argument invalid.\n", argv[1]);
					exit(-1);
				}
			}
			args->pmax = atof(argv[2]);
			if ((args->pmax < XMANDELBROTARGS_MINP) ||
					(args->pmax > XMANDELBROTARGS_MAXP)) {
				fprintf(stderr, "%s: argument out of range.\n", argv[1]);
				exit(-1);
			}
    	--argc;
			++argv;
		}
		else if (strcmp(argv[1], "-qmin") == 0) {
			if ((argc == 2) || ((argv[2][0] == '-') && (!isdigit(argv[2][1])))){
				fprintf(stderr, "%s: argument missing.\n", argv[1]);
				exit(-1);
    	}
			// check for invalid characters:
			for (unsigned int i=0; i<strlen(argv[2]); ++i) {
				if ((!isdigit(argv[2][i])) &&
						(argv[2][i] != '.') &&
						(argv[2][i] != '+') &&
						(argv[2][i] != '-')) {
      		fprintf(stderr, "%s: argument invalid.\n", argv[1]);
					exit(-1);
				}
			}
			args->qmin = atof(argv[2]);
			if ((args->qmin < XMANDELBROTARGS_MINQ) ||
					(args->qmin > XMANDELBROTARGS_MAXQ)) {
				fprintf(stderr, "%s: argument out of range.\n", argv[1]);
				exit(-1);
			}
    	--argc;
			++argv;
		}
		else if (strcmp(argv[1], "-qmax") == 0) {
			if ((argc == 2) || ((argv[2][0] == '-') && (!isdigit(argv[2][1])))){
				fprintf(stderr, "%s: argument missing.\n", argv[1]);
				exit(-1);
    	}
			// check for invalid characters:
			for (unsigned int i=0; i<strlen(argv[2]); ++i) {
				if ((!isdigit(argv[2][i])) &&
						(argv[2][i] != '.') &&
						(argv[2][i] != '+') &&
						(argv[2][i] != '-')) {
      		fprintf(stderr, "%s: argument invalid.\n", argv[1]);
					exit(-1);
				}
			}
			args->qmax = atof(argv[2]);
			if ((args->qmax < XMANDELBROTARGS_MINQ) ||
					(args->qmax > XMANDELBROTARGS_MAXQ)) {
				fprintf(stderr, "%s: argument out of range.\n", argv[1]);
				exit(-1);
			}
    	--argc;
			++argv;
		}
		else if ((strcmp(argv[1], "-t") == 0) ||
							(strcmp(argv[1], "-title") == 0)) {
			if ((argc == 2) || (argv[2][0] == '-')){
				fprintf(stderr, "%s: argument missing.\n", argv[1]);
				exit(-1);
    	}
			strnzcpy(args->title, argv[2], sizeof(args->title));
    	--argc;
			++argv;
		}
		else if ((strcmp(argv[1], "-v") == 0) ||
							(strcmp(argv[1], "-verbose") == 0)) {
			args->verbosity = 1;
		}
		else {
			fprintf(stderr, "%s: unknown argument.\n", argv[1]);
			fprintf(stderr, "(use -help for list)\n");
			exit(-1);
  	}
		--argc;
		++argv;
	}
	return(args);
}
