//+--------------------------------------------------------------------------
//|
//| utilities.cpp - implementation of some helpful functions
//|
//| This file is part of a larger package; see the README supplied with
//| this for information about the package, how to build and how to run
//| it.
//|
//| (c) 2002, 2020 Dietmar Koenig <http://www.faoileag.de>
//|
//| This program is free software: you can redistribute it and/or modify
//| it under the terms of the GNU General Public License as published by
//| the Free Software Foundation, either version 3 of the License, or
//| (at your option) any later version.
//|
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//| GNU General Public License for more details.
//|
//| You should have received a copy of the GNU General Public License
//| along with this program.  If not, see <https://www.gnu.org/licenses/>.
//|
//+--------------------------------------------------------------------------
#include <time.h>
#include "utilities.h"

//+--------------------------------------------------------------------------+
//|                                                                          |
//| strnzcpy(char* dest, char* src, unsigned int n)                          |
//|                                                                          |
//| An extension of the strncpy() function that sets the last byte of the    |
//| destination string to zero to ensure a zero terminated string (if src    |
//| would be larger than n, than the first n bytes of src would be copied    |
//| into dest. If n would be equal to the size of dest, this would result    |
//| in a string without a terminating zero).                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
char* strnzcpy(char* dest, char* src, unsigned int n) {
	memset(dest, 0, n);
	strncpy(dest, src, n);
	dest[n-1] = 0;
	return(dest);
}

//+--------------------------------------------------------------------------+
//|                                                                          |
//| strnzcat(char* dest, char* src, unsigned int n)                          |
//|                                                                          |
//| An altered and extended version of the strncat() function that sets the  |
//| last byte of the destination string to zero to ensure a zero terminated  |
//| string and also makes sure that only so many characters are concatenated |
//| that will not result in a string larger than dest. In this version, n    |
//| therefore takes the size of dest rather than the number of bytes of src  |
//| to concatenate.                                                          |
//|                                                                          |
//+--------------------------------------------------------------------------+
char* strnzcat(char* dest, char* src, unsigned int n) {
	if (strlen(dest) >= n-1)
		return(dest);
	char strbuf[LONGSTRING];
	char* str = (char*)memset(&strbuf, 0, sizeof(strbuf));
	strnzcpy(str, src, sizeof(strbuf));
	while (strlen(dest)+strlen(str) >= n-1)
		str[strlen(str)-1] = 0;
	strcat(dest, str);
	dest[n-1] = 0;
	return(dest);
}

//+--------------------------------------------------------------------------+
//|                                                                          |
//| char* uuidgen(char* uuid)                                                |
//|                                                                          |
//| uuidgen(char* uuid) creates a string which can be used as a universal    |
//| unique identifier by assembling a pseudo random sequence of 16 letters   |
//| or digits (which is equivalent to roughly 4.76e+28 permutations).        |
//| It writes this 16 characters long string into the memory location        |
//| pointed to by the char pointer uuid; it is up to the caller of this      |
//| function to assure that enough memory has been allocated.                |
//| uuidgen() returns a pointer to the memory location holding the uuid.     |
//|                                                                          |
//+--------------------------------------------------------------------------+
char* uuidgen(char* uuid) {
	int randnum;
	for (int i = 0; i < 15; ++i) {
		randnum = (rand()%62) + 48;
		if (randnum >= 84)
			randnum += 13;
		else if (randnum >= 58)
			randnum += 7;
		uuid[i] = (char)randnum;
	}
	return(uuid);
}

//+--------------------------------------------------------------------------+
//|                                                                          |
//| char* gettime(char* timebuf, int sizeoftimebuf, char* format)            |
//|                                                                          |
//| gettime() generates a string holding the current time in the format      |
//| specified.                                                               |
//| gettime() returns a pointer to the memory location holding the time      |
//| string.                                                                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
char* gettime(char* timebuf, int sizeoftimebuf, char* format) {
	memset(timebuf, 0, sizeoftimebuf);
	time_t now = time(NULL);
	struct tm* time = localtime(&now);
	strftime(timebuf, sizeoftimebuf,"%X", time);
	return(timebuf);
}
