//+--------------------------------------------------------------------------
//|
//| xmandelbrot.cpp - displays a mandelbrot set in a X window
//|
//| This file is part of a larger package; see the README supplied with
//| this for information about the package, how to build and how to run
//| it.
//|
//| (c) 2002, 2020 Dietmar Koenig <http://www.faoileag.de>
//|
//| This program is free software: you can redistribute it and/or modify
//| it under the terms of the GNU General Public License as published by
//| the Free Software Foundation, either version 3 of the License, or
//| (at your option) any later version.
//|
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//| GNU General Public License for more details.
//|
//| You should have received a copy of the GNU General Public License
//| along with this program.  If not, see <https://www.gnu.org/licenses/>.
//|
//+--------------------------------------------------------------------------
#include "xmandelbrot.h"
#include <iostream>
#include <iomanip>

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XMandelbrot::XMandelbrot(...) : XWindow(...)                             |
//|                                                                          |
//| The overwritten standard constructor.                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
XMandelbrot::XMandelbrot(int x, int y, int width, int height, char* title,
	 int canvaswidth, int canvasheight,
	 int immediate, int verbosity, char* filename,
	 int infinity, int noapsect, int function,
	 double pmin, double pmax, double qmax, double qmin,
	 int argc, char** argv) :
	XWindow(x, y, width, height, title, argc, argv) {
	// initalize memspace for strings:
	this->filename = (char*)memset(&filenamebuf[0], 0, sizeof(filenamebuf));
	this->progname = (char*)memset(&prognamebuf[0], 0, sizeof(prognamebuf));
	this->color = new XMandelbrotColors();

	// initialize fields for which there exist command line arguments:
	strnzcpy(this->progname, argv[0], sizeof(prognamebuf));
	strnzcpy(this->filename, filename, sizeof(filenamebuf));
	this->infinity = infinity;
	this->immediate = immediate;
	this->verbosity = verbosity;
	this->noaspect = noaspect;
	this->function = function;
	this->canvaswidth = canvaswidth;
	this->canvasheight = canvasheight;
	this->pmin = pmin;
	this->pmax = pmax;
	this->qmax = qmax;
	this->qmin = qmin;

	// if verbosity is requested say something:
	if (this->verbosity > 0) {
		fprintf(stdout, "\nStarting %s...\n", this->progname);
	}

	// if no filename has been specified, set it to a generic one:
	int filenamespecified = 1;
	if (!strcmp(this->filename, "")) {
		filenamespecified = 0;
		strnzcpy(this->filename, this->progname, sizeof(filenamebuf));
		strnzcat(this->filename, ".saves", sizeof(filenamebuf));
		if (this->verbosity > 0) {
			fprintf(stdout, "No filename specified, using generic.\n");
		}
	}

	// initialize some internal fields:
	strnzcpy(this->progname, argv[0], sizeof(prognamebuf));
	this->kmax = 278; // only use subset of colorscheme
	this->fx = 0;     // first corner of zooming rect x
	this->fx = 0;     // first corner of zooming rect y
	this->sx = 0;     // second corner of zooming rect x
	this->sy = 0;     // second corner of zooming rect y
	this->dp = 0;     // delta p
	this->dq = 0;     // delta q
	this->action = 0; // if 1 do something in event loop
	this->recton = 0; // if 1 zooming rectangle visible

	// set the canvas:
	this->canvasleft = 0;     // offset where we start painting, x
	this->canvastop = 0;      // "      "     "  "     "       , y
	this->cx = 0;
	this->cy = 0;
	if (this->canvaswidth > 0) {
		if (this->canvaswidth < this->width) {
			this->canvasleft = (this->width - this->canvaswidth)/2;
		}
		if (this->canvasheight < this->height) {
			this->canvastop = (this->height - this->canvasheight)/2;
		}
	}
	else {
		this->canvaswidth = this->width;
		this->canvasheight = this->height;
	}

	// if immediate is set, trigger the start of drawing :
	if (this->immediate) {
		if (filenamespecified)
			this->Load();
		this->Aspect();
		this->dp = (this->pmax - this->pmin) / this->canvaswidth;
		this->dq = (this->qmax - this->qmin) / this->canvasheight;
		this->action = 1;
	}

	// if wanted, print some feedback:
	if (this->verbosity > 0) {
		fprintf(stdout, "Initial values are:\n");
		fprintf(stdout, "Width           = %4d\n", this->width);
		fprintf(stdout, "Height          = %4d\n", this->height);
		fprintf(stdout, "Canvas width    = %4d\n", this->canvaswidth);
		fprintf(stdout, "Canvas height   = %4d\n", this->canvasheight);
		fprintf(stdout, "Canvas offset x = %4d\n", this->canvasleft);
		fprintf(stdout, "Canvas offset y = %4d\n", this->canvastop);
		fprintf(stdout, "pmin            = %+1.10f\n", this->pmin);
		fprintf(stdout, "pmax            = %+1.10f\n", this->pmax);
		fprintf(stdout, "qmin            = %+1.10f\n", this->qmin);
		fprintf(stdout, "qmax            = %+1.10f\n", this->qmax);
		if (this->verbosity > 1) {
			fprintf(stdout, "delta p         = %+1.10f\n", this->dp);
			fprintf(stdout, "delta q         = %+1.10f\n", this->dq);
		}
		fprintf(stdout, "\n");
		fprintf(stdout, "Infinity is set to %4d,\n", this->infinity);
		fprintf(stdout, "colordepths is %4d and\n", this->kmax);
		fprintf(stdout, "and aspect maintenance is ");
		if (this->noaspect)
			fprintf(stdout, "off.\n");
		else
			fprintf(stdout, "on.\n");
		fprintf(stdout, "File used for loading/saving\n");
		fprintf(stdout, "is %s.\n", this->filename);
		fprintf(stdout, "\n");
	}
} // end of XMandelbrot::XMandelbrot(...) : XWindow(...)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| virtual void XMandelbrot::EventLoopAction()                              |
//|                                                                          |
//| Calculate the next pixel in the mandelbrot set.                          |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::EventLoopAction() {
	if (this->action == 0) {
		return;
	}
	this->DrawPixel(this->cx, this->cy);
	++this->cx;
	if (this->cx == this->canvaswidth) {
		this->cx = 0;
		++this->cy;
		if (this->cy > this->canvasheight) {
			this->action = 0;
			if (this->verbosity > 0) {
				fprintf(stdout, "Finished painting.\n");
			}
		}
  }
} // end of XMandelbrot::EventLoopAction()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| virtual void XMandelbrot::EvHandlerButton1Press(XButtonEvent* event)     |
//|                                                                          |
//| Pressing the left mouse button sets the upper left corner of a region    |
//| selecting rectangle.                                                     |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::EvHandlerButton1Press(XButtonEvent* event) {
  if (this->verbosity > 1) {
	  fprintf(stdout, "left mouse button clicked\n");
	}
	// new rectangle - delete any previous one:
	if (this->recton) {
    if (this->verbosity > 1) {
	    fprintf(stdout, "deleting old rect\n");
  	}
		this->RemoveRect(this->fx, this->fy, this->sx, this->sy);
  }

	// are coordinates on canvas?
	if ((event->x < this->canvasleft) ||
			(event->y < this->canvastop) ||
			(event->x > (this->canvasleft + this->canvaswidth)) ||
			(event->y > (this->canvastop + this->canvasheight))) {
		this->fx = 0;
		this->fy = 0;
		this->sx = 0;
		this->sy = 0;
		if (this->verbosity > 0) {
			fprintf(stdout, "outside of canvas.\n");
		}
		return;
	}

	// set coordinates and draw new rectangle:
	this->fx = event->x;
	this->sx = event->x;
	this->fy = event->y;
	this->sy = event->y;
	this->DrawRect(this->fx, this->fy, this->sx, this->sy);
	if (this->verbosity > 1) {
		fprintf(stdout, "rectangle: left %4d top %4d right %4d bottom %4d\n",
			this->fx, this->fy, this->sx, this->sy);
  }
} // end of XMandelbrot::EvHandlerButton1Press(XButtonEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| virtual void XMandelbrot::EvHandlerButton1MotionNotify(XButtonEvent*...) |
//|                                                                          |
//| delete the previous rectangle then draw a new one.                       |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::EvHandlerButton1MotionNotify(XButtonEvent* event) {
	// if no rectangle is visible, nothing happens:
	if (!this->recton)
		return;

	// are new coordinates still on canvas?
	if ((this->fx < this->canvasleft) || (this->fy < this->canvastop)) {
		if (this->verbosity > 0) {
			fprintf(stdout, "outside of canvas.\n");
		}	
		return;
  }	
	if ((event->x > (this->canvasleft + this->canvaswidth)) ||
			(event->y > (this->canvastop + this->canvasheight))) {
		if (this->verbosity > 0) {
			fprintf(stdout, "outside of canvas.\n");
		}
		return;
	}
	if ((event->x < this->canvasleft) || (event->y < this->canvastop)) {
		if (this->verbosity > 0) {
			fprintf(stdout, "outside of canvas.\n");
		}
		return;
	}

	// delete any previous rectangle:
  if (this->verbosity > 1) {
	  fprintf(stdout, "deleting previous rect\n");
	}
	this->RemoveRect(this->fx, this->fy, this->sx, this->sy);
	this->sx = event->x;
	this->sy = event->y;

	// draw new rectangle:
  if (this->verbosity > 1) {
	  fprintf(stdout, "drawing new rect\n");
	}
	this->DrawRect(this->fx, this->fy, this->sx, this->sy);
	if (this->verbosity > 1) {
		fprintf(stdout, "Rectangle: left %4d top %4d right %4d bottom %4d\n",
			this->fx, this->fy, this->sx, this->sy);
  }
}

//+--------------------------------------------------------------------------+
//|                                                                          |
//| virtual void XMandelbrot::EvHandlerButton1Release(XButtonEvent* event    |
//|                                                                          |
//| draws the inverted rectangle at the last position.                       |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::EvHandlerButton1Release(XButtonEvent* event) {
	// if no rectangle is visible, nothing happens:
	if (!this->recton)
		return;

	// just some mouse click into the window?
	if ((event->x == this->fx) && (event->y == this->fy)) {
		this->RemoveRect(this->fx, this->fy, this->fx, this->fy);
		this->fx = 0;
		this->fy = 0;
		this->sx = 0;
		this->sy = 0;
		return;
	}

	// are new coordinates still on canvas?
	if ((this->fx < this->canvasleft) || (this->fy < this->canvastop)) {
		if (this->verbosity > 0) {
			fprintf(stdout, "Outside of canvas.\n");
		}
		return;
	}
	if ((event->x > (this->canvasleft + this->canvaswidth)) ||
			(event->y > (this->canvastop + this->canvasheight))) {
		if (this->verbosity > 0) {
			fprintf(stdout, "Outside of canvas.\n");
		}
		return;
	}
	if ((event->x < this->canvasleft) || (event->y < this->canvastop)) {
		if (this->verbosity > 0) {
			fprintf(stdout, "Outside of canvas.\n");
		}
		return;
	}

	// delete any previous rectangle:
	this->RemoveRect(this->fx, this->fy, this->sx, this->sy);
	this->sx = event->x;
	this->sy = event->y;

	// draw new rectangle:
	this->DrawRect(this->fx, this->fy, this->sx, this->sy);
	if (this->verbosity > 0) {
		fprintf(stdout, "Selected area is:\n");
		fprintf(stdout, "left %4d top %4d right %4d bottom %4d\n",
			this->fx, this->fy, this->sx, this->sy);
  }
}

//+--------------------------------------------------------------------------+
//|                                                                          |
//| virtual void XMandelbrot::EvHandlerButton2Release(XButtonEvent* event)   |
//|                                                                          |
//| clicking the middle mouse button closes the window.                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::EvHandlerButton2Release(XButtonEvent* event) {
	this->Quit();
}

//+--------------------------------------------------------------------------+
//|                                                                          |
//| virtual void XMandelbrot::EvHandlerButton3Release(XButtonEvent* event)   |
//|                                                                          |
//| clicking the right mouse button initiates the drawing of a new set.      |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::EvHandlerButton3Release(XButtonEvent* event) {
	this->Zoom();
	this->Start();
}

//+--------------------------------------------------------------------------+
//|                                                                          |
//| virtual void XMandelbrot::EvHandlerKeyPress(XKeyEvent* event)            |
//|                                                                          |
//| finds out which key has been pressed and acts apropriately.              |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::EvHandlerKeyPressed(XKeyEvent* event) {
	char* keystr =
		XKeysymToString(XKeycodeToKeysym(this->display, event->keycode, 0));
	if (this->verbosity > 1) {
		fprintf(stdout, "Key pressed \"\%s\"\n", keystr);
  }
	if (strcmp(keystr, "Escape") == 0) {
		if (this->recton)
			this->RemoveRect(this->fx, this->fy, this->sx, this->sy);
		this->fx = 0;
		this->fy = 0;
		this->sx = 0;
		this->sy = 0;
		if (this->verbosity > 0) {
			fprintf(stdout, "Cleared selected area.\n");
	  }
  }
  else if (strcmp(keystr, "d") == 0) {
		this->Zoom();
		this->Start();
  }
  else if (strcmp(keystr, "l") == 0) {
		this->Load();
		this->Start();
	}
  else if (strcmp(keystr, "p") == 0) {
		fprintf(stdout, "Current paramters are:\n");
		fprintf(stdout, "pmin = %+1.10f pmax = %+1.10f\n", this->pmin, this->pmax);
		fprintf(stdout, "qmin = %+1.10f qmax = %+1.10f\n", this->qmin, this->qmax);
		fprintf(stdout, "width: %4d height: %4d\n", this->width, this->height);
		fprintf(stdout, "\n");
	}
	else if (strcmp(keystr, "q") == 0)
		this->Quit();
  else if (strcmp(keystr, "s") == 0)
		this->Save();
}

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMandelbrot::DrawPixel(int x, int y)                                |
//|                                                                          |
//| Calculates the value of the pixel specified by x nd y.                   |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::DrawPixel(int x, int y) {
	unsigned long pixelcolor = BlackPixel(this->display, this->screen);
	switch (this->function) {
	case XMANDELBROTARGS_FUNCTIONMANDELBROT:
		pixelcolor = this->Mandelbrot(x, y);
		break;
	case XMANDELBROTARGS_FUNCTIONJULIA:
		pixelcolor = this->Julia(x, y);
		break;
	default:
		pixelcolor = this->Mandelbrot(x, y);
	}
	XDrawPoint(this->display, this->videomem, this->gc, this->canvasleft + x, this->canvastop + y);
 	XSetForeground(this->display, this->gc, pixelcolor);
 	XDrawPoint(this->display, this->window, this->gc, this->canvasleft + x, this->canvastop + y);
} // end of void XMandelbrot::DrawPixel(int x, int y)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMandelbrot::DrawRect(int c1x, int c1y, int c2x, int c2y)           |
//|                                                                          |
//| Draws a rectangle using the two given corners.                           |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::DrawRect(int c1x, int c1y, int c2x, int c2y) {
	if (this->verbosity > 1) {
		std::cout << "drawing rect " << c1x << "," << c1y << ",";
    std::cout << c2x << "," << c2y << "\n";
  }

	// determine which corner is top left:
	if (c2x < c1x) {
		int xbuf = c1x;
		c1x = c2x;
		c2x = xbuf;
	}
	if (c2y < c1y) {
		int ybuf = c1y;
		c1y = c2y;
		c2y = ybuf;
	}

	// draw horizontal lines:
  for (int i = 0; i < (c2x-c1x); i++) {
    int cxi = c1x+i;
	  XSetForeground(this->display, this->gc,
	    BlackPixel(this->display, this->screen));
 	 	XDrawPoint(this->display, this->window, this->gc, cxi, c1y);
 	  XDrawPoint(this->display, this->window, this->gc, cxi, c2y);
  }
  // draw vertical lines:
  for (int i = 0; i < (c2y-c1y); i++) {
   	int cyi = c1y+i;
    XSetForeground(this->display, this->gc,
      BlackPixel(this->display, this->screen));
 	 	XDrawPoint(this->display, this->window, this->gc, c1x, cyi);
 	  XDrawPoint(this->display, this->window, this->gc, c2x, cyi);
  }
	this->recton = 1;
	if (this->verbosity > 1) {
    std::cout << "done drawing the rectangle\n";
  }
} // end of void XMandelbrot::DrawRect(int c1x, int c1y, int c2x, int c2y)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMandelbrot::RemoveRect(int c1x, int c1y, int c2x, int c2y)         |
//|                                                                          |
//| Removes an existing rectangle by repainting the area from videomem.      |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::RemoveRect(int c1x, int c1y, int c2x, int c2y) {
	if (this->verbosity > 1) {
		std::cout << "deleting rect " << c1x << "," << c1y << ",";
    std::cout << c2x << "," << c2y << "\n";
  }

	// determine which corner is top left:
	if (c2x < c1x) {
		int xbuf = c1x;
		c1x = c2x;
		c2x = xbuf;
	}
	if (c2y < c1y) {
		int ybuf = c1y;
		c1y = c2y;
		c2y = ybuf;
	}

  int left   = c1x;
  int top    = c1y;
  int width  = c2x - c1x + 1;
  int height = c2y - c1y + 1;

	// delete rect by repainting area from videomem:
  XCopyArea(this->display, this->videomem, this->window,
    this->gc, left, top, width, height, left, top);

	this->recton = 0;

	if (this->verbosity > 1) {
    std::cout << "done deleting the rectangle\n";
  }
}

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMandelbrot::Aspect()                                               |
//|                                                                          |
//| Pads the boundaries to maintain aspect ratio of image against that of    |
//| the canvas.                                                              |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::Aspect() {
	//if aspect maintenance is not wanted, just return immediately:
	if (this->noaspect)
		return;

	// determine actual and needed aspect ratio:
	double pqplanear = (this->pmax-this->pmin)/(this->qmax-this->qmin);
	double canvasar = ((double)this->canvaswidth) / ((double)this->canvasheight);
	if (this->verbosity > 1) {
		fprintf(stderr, "pq-plane aspect ratio: %1.12f.\n", pqplanear);
		fprintf(stderr, "canvas   aspect ratio: %1.12f.\n", canvasar);
	}
	if (pqplanear == canvasar)
		return;

	// padding needed - pad the p-axis:
	double pspread = (this->qmax-this->qmin) * canvasar;
	double padding = (pspread - (this->pmax-this->pmin)) / 2;
	this->pmin -= padding;
	this->pmax += padding;
	if (this->verbosity > 1) {
		fprintf(stderr, "pmin is now: %1.12f.\n", this->pmin);
		fprintf(stderr, "pmax is now: %1.12f.\n", this->pmax);
	}

	// if verbosity is wanted, say something:
	if (this->verbosity > 0) {
		fprintf(stdout, "P-axis has been padded by %1.12f.\n", padding*2);
	}
} // end of void XMandelbrot::Aspect()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMandelbrot::Zoom()                                                 |
//|                                                                          |
//| calculates the new boundaries.                                           |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::Zoom() {
	if (!this->recton)
		return;
	this->RemoveRect(this->fx, this->fy, this->sx, this->sy);
	if ((this->sx == this->fx) || (this->sy == this->fy))
		return;

	// determine which corner is top left:
	if (this->sx < this->fx) {
		int xbuf = this->fx;
		this->fx = this->sx;
		this->sx = xbuf;
	}
	if (this->sy < this->fy) {
		int ybuf = this->fy;
		this->fy = this->sy;
		this->sy = ybuf;
	}

	// calculate new boundaries:
	this->pmax = this->pmin + (this->dp * (this->sx - this->canvasleft));
	this->pmin = this->pmin + (this->dp * (this->fx - this->canvasleft));
	this->qmin = this->qmax - (this->dq * (this->sy - this->canvastop));
	this->qmax = this->qmax - (this->dq * (this->fy - this->canvastop));

	// reset area corners:
	this->fx = 0;
	this->fy = 0;
	this->sx = 0;
	this->sy = 0;
} // end of void XMandelbrot::Zoom()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMandelbrot::Start()                                                |
//|                                                                          |
//| initiates the drawing of a new image.                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::Start() {
	// before deltas are calculated, check aspect ratio:
	this->Aspect();

	// calculate deltas:
	this->dp = (this->pmax - this->pmin) / this->canvaswidth;
	this->dq = (this->qmax - this->qmin) / this->canvasheight;

	// clear the videomemeory and the canvas:
	if (this->verbosity > 1) {
    std::cout << "clearing videomem and canvas" << std::endl;
  }
 	XSetForeground(this->display, this->gc,
		WhitePixel(this->display, this->screen));
  XFillRectangle(this->display, this->videomem, this->gc, 0, 0, 1280, 1024);
  XFillRectangle(this->display, this->window, this->gc, 0, 0, this->width, this->height);

	// reset canvas coordinates:
	this->action = 1;
	this->cx = 0;
	this->cy = 0;

	// if wanted, print some feedback:
	if (this->verbosity > 0) {
    std::cout << std::setiosflags(std::ios::fixed);
    std::cout << std::setprecision(6);
		std::cout << "drawing new image with params:" << std::endl;
    std::cout << "  pmin    = " << this->pmin << std::endl;
		std::cout << "  pmax    = " << this->pmax << std::endl;
    std::cout << "  qmin    = " << this->qmin << std::endl;
		std::cout << "  qmax    = " << this->qmax << std::endl;
		if (this->verbosity > 1) {
			std::cout << "  delta p = " << this->dp << std::endl;
			std::cout << "  delta q = " << this->dq << std::endl;
		}
		std::cout << std::endl;
  }
} // end of void XMandelbrot::Start()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMandelbrot::Save()                                                 |
//|                                                                          |
//| Dumps settings and videomemory to file.                                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::Save() {
	// if wanted, print some feedback:
	if (this->verbosity > 0) {
		fprintf(stdout, "Saving current params:\n");
		fprintf(stdout, "pmin = %+1.10f\n", this->pmin);
		fprintf(stdout, "pmax = %+1.10f\n", this->pmax);
		fprintf(stdout, "qmin = %+1.10f\n", this->qmin);
		fprintf(stdout, "qmax = %+1.10f\n", this->qmax);
		fprintf(stdout, "\n");
  }

	// save params to file:
	FILE* file = fopen(this->filename, "w");
	if (file == NULL) {
		fprintf(stderr, "Error writing file: %s\n", filename);
		return;
	}
	fprintf(file, "%+1.10f/%+1.10f/%+1.10f/%+1.10f\n",
					 this->pmin, this->pmax, this->qmin, this->qmax);
	fclose(file);
	if (this->verbosity > 0) {
		fprintf(stdout, "Done.\n");
  }
} // end of void XMandelbrot::Save()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMandelbrot::Load()                                                 |
//|                                                                          |
//| Loads settings and videomemory from file and displays the image.         |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::Load() {
	if (this->verbosity > 0)
		fprintf(stdout, "Loading params...\n");

	// load data from file:
	FILE* file = fopen(this->filename, "r");
	if (file == NULL) {
		fprintf(stderr, "Error reading file: %s\n", this->filename);
		return;
	}
	char line[255];
	char* params = fgets(line, sizeof(line), file);
	fclose(file);

	// parse params from data string:
	if (params == NULL) {
		fprintf(stderr, "Error parsing file: %s\n", this->filename);
		return;
	}
	double pmin, pmax, qmax, qmin;
	char* param = strtok(params, "/");
	if (param == NULL) {
		fprintf(stderr, "Error parsing file: %s\n", this->filename);
		return;
	}
	pmin = atof(param);
	param = strtok(NULL, "/");
	if (param == NULL) {
		fprintf(stderr, "Error parsing file: %s\n", this->filename);
		return;
	}
	pmax = atof(param);
	param = strtok(NULL, "/");
	if (param == NULL) {
		fprintf(stderr, "Error parsing file: %s\n", this->filename);
		return;
	}
	qmin = atof(param);
	param = strtok(NULL, "/");
	if (param == NULL) {
		fprintf(stderr, "Error parsing file: %s\n", this->filename);
		return;
	}
	qmax = atof(param);

	// sanity check:
	if ((pmin < XMANDELBROTARGS_MINP) || (pmin > XMANDELBROTARGS_MAXP) ||
			(pmax < XMANDELBROTARGS_MINP) || (pmax > XMANDELBROTARGS_MAXP) ||
			(qmax < XMANDELBROTARGS_MINQ) || (qmax > XMANDELBROTARGS_MAXQ) ||
			(qmin < XMANDELBROTARGS_MINQ) || (qmin > XMANDELBROTARGS_MAXQ) ||
			(pmin >= pmax) || (qmin >= qmax)) {
		fprintf(stderr, "%s: param out of range.\n", this->filename);
 	}

	// params seem to be ok so overwrite the respective fields:
	this->pmin = pmin;
	this->pmax = pmax;
	this->qmin = qmin;
	this->qmax = qmax;

	if (this->verbosity > 0) {
		fprintf(stdout, "Done.\n");
  }
} // end of void XMandelbrot::Load()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMandelbrot::Quit()                                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMandelbrot::Quit() {
} // end of void XMandelbrot::Quit() {

//+--------------------------------------------------------------------------+
//|                                                                          |
//| unsigned long XMandelbrot::Mandelbrot(int x, int y)                      |
//|                                                                          |
//| Calculates the value of pixel(x,y) in a Mandelbrot Set.                  |
//| Returns the pixel's color as a long value.                               |
//|                                                                          |
//+--------------------------------------------------------------------------+
unsigned long XMandelbrot::Mandelbrot(int x, int y) {
	XColor rgb;
	unsigned long retcol = BlackPixel(this->display, this->screen);
	double p = this->pmin + x*this->dp;
	double q = this->qmax - y*this->dq;
	double r = 0;						// the result
	double xk = 0;					// intermediate value for x
	double yk = 0;					// intermediate value for y
	double bx = 0;					// buffer for xk
	int k = 1;							// the color depth
	int doloop = 1;					// if 1 continue increasing k
	while(doloop) {
		bx = xk;
		xk = (xk*xk) - (yk*yk) + p;
		yk = 2*bx*yk + q;
		++k;
		r = (xk*xk) + (yk*yk);
		if (r > infinity) {
			rgb.red   = this->color->red[k];
			rgb.green = this->color->green[k];
			rgb.blue  = this->color->blue[k];
			XAllocColor(this->display, colormap, &rgb);
			retcol = rgb.pixel;
			//retcol = WhitePixel(this->display, this->screen);
			doloop = 0;
   		}
		if (k == kmax) {
			retcol = BlackPixel(this->display, this->screen);
			doloop = 0;
   		}
  	}
	return(retcol);
} // end of unsigned long XMandelbrot::Mandelbrot(int x, int y)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| unsigned long XMandelbrot::Julia(int x, int y)                           |
//|                                                                          |
//| Calculates the value of pixel(x,y) in a Julia Set.                       |
//| Returns the pixel's color as a long value.                               |
//|                                                                          |
//+--------------------------------------------------------------------------+
unsigned long XMandelbrot::Julia(int x, int y) {
	return(BlackPixel(this->display, this->screen));
} // end of unsigned long XMandelbrot::Julia(int x, int y)

