//+--------------------------------------------------------------------------
//|
//| xwindow.cpp - implementation of various x window related classes
//|
//| This file is part of a larger package; see the README supplied with
//| this for information about the package, how to build and how to run
//| it.
//|
//| (c) 2002, 2020 Dietmar Koenig <http://www.faoileag.de>
//|
//| This program is free software: you can redistribute it and/or modify
//| it under the terms of the GNU General Public License as published by
//| the Free Software Foundation, either version 3 of the License, or
//| (at your option) any later version.
//|
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//| GNU General Public License for more details.
//|
//| You should have received a copy of the GNU General Public License
//| along with this program.  If not, see <https://www.gnu.org/licenses/>.
//|
//+--------------------------------------------------------------------------
#include <time.h>
#include "xwindow.h"
#include "utilities.h"

//+--------------------------------------------------------------------------+
//|                                                                          |
//| function pred(Display* disp, XEvent* event, XPointer arg)                |
//|                                                                          |
//| the predicate function for XCheckIfEvent(). does nothing except          |
//| returning true.                                                          |
//|                                                                          |
//+--------------------------------------------------------------------------+
int pred(Display* disp, XEvent* event, XPointer arg) {
	return(1);
} // end of int pred()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XList::XList()                                                           |
//|                                                                          |
//| Initializes the pointers to NULL and the count to zero.                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
XList::XList() {
	this->cursor = NULL;
	this->first = NULL;
	this->last = NULL;
	this->count =  0;
} // end of XList::XList()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XList::~XList()                                                          |
//|                                                                          |
//| Frees the memory used by the items in the list.                          |
//|                                                                          |
//+--------------------------------------------------------------------------+
XList::~XList() {
	Clear();
} // end of XList::~XList()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XList::Add(XControl* item)                                           |
//|                                                                          |
//| Adds the item specified as argument to the list. After execution of this |
//| method the cursor points to the newly added item.                        |
//| Returns 0 if adding fails, otherwise 1.                                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XList::Add(XControl* item) {
	// Allocate memory for list item:
	TList* newelement = (TList*)malloc(sizeof(TList));
	newelement->control = item;
	newelement->next = NULL;
	newelement->prev = this->last;
	if (this->first == NULL)
		this->first = newelement;
	if (this->last != NULL)
		this->last->next = newelement;
	this->last = newelement;
	this->cursor = this->last;
	++this->count;
	return(1);
} // end of int XList::Add(XControl* item)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XList::Insert(XControl* item)                                        |
//|                                                                          |
//| Insert the item specified as argument before the item the cursor points  |
//| to. After execution of this method the cursor points to the inserted     |
//| item.                                                                    |
//| Returns 0 if inserting fails, otherwise 1.                               |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XList::Insert(XControl* item) {
	// Allocate memory for new list item:
	TList* newitem = (TList*)malloc(sizeof(TList));
	newitem->control = item;
	newitem->next = NULL;
	newitem->prev = NULL;
	if (this->count == 0) {
		this->first = newitem;
		this->last = this->first;
		this->cursor = this->first;
	}
	else {
		newitem->next = this->cursor;
		newitem->prev = this->cursor->prev;
		this->cursor->prev = newitem;
	}
	this->cursor = newitem;
	++this->count;
	return(1);
} // end of int XList::Insert(XControl* item)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XList::Remove(XControl* item)                                        |
//|                                                                          |
//| Removes the item specified as argument from the list. After execution of |
//| this method the cursor points to the item next in list to the removed    |
//| one if the item was found and removed, otherwise to NULL.                |
//| Returns 0 if removing fails, otherwise 1.                                |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XList::Remove(XControl* item) {
	// Sanitary check:
	if (this->cursor == NULL)
		return(0);
	if (item == NULL)
		return(0);

	// Scan the list for the item and if found remove it:
	TList* nextbuf = NULL;
	TList* prevbuf = NULL;
	this->cursor = this->first;
	while(this->cursor != NULL) {
		nextbuf = this->cursor->next;
		if (this->cursor->control == item) {
			// Found item. Remove it:
			free(this->cursor);
			if (prevbuf == NULL) {
				if (nextbuf != NULL)
					nextbuf->prev = NULL;
				this->first = nextbuf;
			}
			else {
				if (nextbuf != NULL)
					nextbuf->prev = prevbuf;
				prevbuf->next = nextbuf;
			}
			this->cursor = nextbuf;
			--this->count;
			return(1);
		}
		prevbuf = this->cursor;
		this->cursor = this->cursor->next;
	}
	return(0);
} // end of int XList::Remove(XControl* item)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XList::Clear()                                                       |
//|                                                                          |
//| Removes all items from the list and frees the memory of the items. All   |
//| pointers are set to NULL and count to 0.                                 |
//| Returns 0 if clearing fails, otherwise 1.                                |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XList::Clear() {
	// Check if list already is empty:
	if (this->count == 0)
		return(1);

	// Clear the list item by item:
	while(this->first != NULL) {
		this->cursor = this->first->next;
		free(this->first);
		this->first = this->cursor;
	}
	this->last = NULL;
	this->cursor = NULL;
	this->count = 0;
	return(1);
} // end of int XList::Clear()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XList::Reset()                                                       |
//|                                                                          |
//| Sets the cursor to the first item in the list.                           |
//| Returns 1.                                                               |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XList::Reset() {
	this->cursor = this->first;
	return(1);
} // end of int XList::Reset()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XList::Next()                                                        |
//|                                                                          |
//| Moves the cursor to the next item in the list.                           |
//| Returns 0 if cursor is pointing to NULL, otherwise 1.                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XList::Next() {
	if (this->cursor == NULL)
		return(0);
	this->cursor = this->cursor->next;
	return(1);
} // end of int XList::Next()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XList::Previous()                                                    |
//|                                                                          |
//| Moves the cursor to the previous item in the list.                       |
//| Returns 0 if cursor is pointing to NULL, otherwise 1.                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XList::Previous() {
	if (this->cursor == NULL)
		return(0);
	this->cursor = this->cursor->prev;
	return(1);
} // end of int XList::Previous()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XControl* XList::Item()                                                  |
//|                                                                          |
//| Returns NULL or a pointer to the control the cursor points at.           |
//|                                                                          |
//+--------------------------------------------------------------------------+
XControl* XList::Item() {
	if (this->cursor == NULL)
		return(NULL);
	return(this->cursor->control);
} // end of XControl* XList::Item()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XList::Count()                                                       |
//|                                                                          |
//| Returns this->count.                                                     |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XList::Count() {
	return(this->count);
} // end of int XList::Count()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XTabRing::XTabRing() : XList()                                           |
//|                                                                          |
//| All initialization is done in the constructor of the base class.         |
//|                                                                          |
//+--------------------------------------------------------------------------+
XTabRing::XTabRing() : XList() {
} // end of XTabRing::XTabRing() : XList()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XTabRing::Insert(XControl* item)                                     |
//|                                                                          |
//| Sorts the control specified by the argument into the tab ring. The       |
//| cursor is not affected.                                                  |
//| Returns 1 is the control could succesfully be sorted into the tab ring   |
//| and also if the control does not allow to be tab-ringed (i.e. by setting |
//| taborder to XW_NOTABRING).                                               |
//| Returns 1 if the control could succesfully be sorted into the tab ring   |
//| and 0 if the control is of a not accepted type.                          |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XTabRing::Insert(XControl* item) {
	// Sanity check:
	if (item == NULL)
		return(0);

	// Some controls are not accepted in the tabring:
	if ((item->Type() == XW_TYPESTATUSBAR) ||
			(item->Type() == XW_TYPESTATUSBARITEM))
		return(0);

	// Get items tab order:
	int taborder = item->TabOrder();

	// If control is not to be "tab ringed" jump to next control:
	if (taborder == XW_NOTABRING)
		return(1);

	// "Split" the loop:
	if (this->last != NULL)
		this->last->next = NULL;
	if (this->first != NULL)
		this->first->prev = NULL;

	// Scan tab ring for a control with a higher taborder and insert before:
	TList* newitem = (TList*)malloc(sizeof(TList));
	TList* ptr = NULL;
	ptr = this->first;
	while (ptr != NULL) {
		if (ptr->control->TabOrder() > taborder) {
			newitem->control = item;
			newitem->next = ptr;
			newitem->prev = ptr->prev;
			if (ptr == this->first) {
				this->first->prev = newitem;
				this->first = newitem;
			}
			else {
				ptr->prev->next = newitem;
				ptr->prev = newitem;
			}
			break;
		}
		ptr = ptr->next;
	}

	// If ptr == NULL no control has been found -> add item as last:
	if (ptr == NULL) {
		newitem->control = item;
		newitem->next = NULL;
		newitem->prev = this->last;
		if (this->last != NULL)
			this->last->next = newitem;
		this->last = newitem;
		if (this->first == NULL)
			this->first = newitem;
	}

	// Restore loopness:
	this->last->next = this->first;
	this->first->prev = this->last;

	// Set the cursor in case this has not yet happened:
	if (this->cursor == NULL)
		this->cursor = this->first;

	// And finally increase count:
	++count;
	return(1);
} // end of int XTabRing::Insert(XControl* item)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XTabRing::Remove(XControl* item)                                     |
//|                                                                          |
//| Removes the control specified as argument from the list. After succesful |
//| removal the cursor points to the control next in list to the removed one.|
//| If the control is focused it is set back to normal before it is removed. |
//| Returns 1 if teh control could be removed successfully or was not in the |
//| tab ring and 0 a null pointer was passed as control.                     |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XTabRing::Remove(XControl* item) {
	// Sanity check:
	if (this->count == 0)
		return(0);

	// "Split" the loop:
	this->last->next = NULL;
	this->first->prev = NULL;

	// Scan all controls for the one to be removed:
	TList* nextbuf = NULL;
	TList* prevbuf = NULL;
	this->cursor = this->first;
	while(this->cursor != NULL) {
		if (this->cursor->control == item) {
			if (this->cursor->control->Focused())
				this->cursor->control->SetNormal();
			nextbuf = this->cursor->next;
			if (this->cursor == this->first)
				this->first = nextbuf;
			if (this->cursor == this->last)
				this->last = prevbuf;
			free(this->cursor);
			if (nextbuf != NULL)
				nextbuf->prev = prevbuf;
			if (prevbuf != NULL)
				prevbuf->next = nextbuf;
			this->cursor = nextbuf;
			--count;
			break;
		}
		prevbuf	= this->cursor;
		this->cursor = this->cursor->next;
	}

	// Restore loopness:
	if (this->last != NULL)
		this->last->next = this->first;
	if (this->first != NULL)
		this->first->prev = this->last;
	return(1);
} // end of int XTabRing::Remove(XControl* item)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XTabRing::SetFocusTo(XControl* item)                                 |
//|                                                                          |
//| Sets the focus and the cursor to the control specified in the argument.  |
//| All other controls are set to normal.                                    |
//| Returns 1 if the operation was succesful, 0 in case of failure.          |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XTabRing::SetFocusTo(XControl* item) {
	// Sanity check:
	if (this->count == 0)
		return(0);

	// "Split" the loop:
	this->last->next = NULL;
	this->first->prev = NULL;

	// Scan all controls for the one to be focused:
	int found = 0;
	TList* ptr = this->first;
	while(ptr != NULL) {
		if (ptr->control == item) {
			ptr->control->SetFocused();
			this->cursor = ptr;
			found = 1;
		}
		else
			ptr->control->SetNormal();
		ptr = ptr->next;
	}

	// Restore loopness:
	if (this->last != NULL)
		this->last->next = this->first;
	if (this->first != NULL)
		this->first->prev = this->last;

	// Determine outcome:
	if (!found)
		return(0);
	return(1);
} // end of int XTabRing::SetFocusTo(XControl* item)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XTabRing::Clear()                                                    |
//|                                                                          |
//| Sets the current control back to normal, then removes all controls from  |
//| the tab ring and frees the memory used by them. All pointers are set to  |
//| NULL and count to 0.                                                     |
//| Returns 0 if clearing fails, otherwise 1.                                |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XTabRing::Clear() {
	// Sanity Check:
	if (this->cursor == NULL)
		return(0);

	// "Split" the loop:
	if (this->last != NULL)
		this->last->next = NULL;

	// Now we can call XList's Clear()
	if (!XList::Clear())
		return(0);
	return(1);
} // end of int XTabRing::Clear()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XTabRing::Reset()                                                    |
//|                                                                          |
//| Sets the cursor to the first control in the tab ring.                    |
//| Returns 0 if cursor is pointing to NULL, otherwise 1.                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XTabRing::Reset() {
	// Sanity Check:
	if (this->cursor == NULL)
		return(0);

	// Set focus of current control back to normal:
	this->cursor->control->SetNormal();

	// Move the cursor:
	this->cursor = this->first;

	// 2. Sanity check:
	if (this->cursor == NULL)
		return(0);

	// Set the current control as focused:
	this->cursor->control->SetFocused();
	return(1);
} // end of int XTabRing::Reset()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XTabRing::Next()                                                     |
//|                                                                          |
//| Sets the current control back to normal, moves the cursor to the next    |
//| control in the tab ring and sets that to focused.                        |
//| Returns 0 if cursor is pointing to NULL, otherwise 1.                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XTabRing::Next() {
	// Sanity Check:
	if (this->cursor == NULL)
		return(0);

	// Set focus of current control back to normal:
	this->cursor->control->SetNormal();

	// Move the cursor:
	this->cursor = this->cursor->next;

	// 2. Sanity check:
	if (this->cursor == NULL)
		return(0);

	// Set the current control as focused:
	this->cursor->control->SetFocused();
	return(1);
} // end of int XTabRing::Next()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XTabRing::Previous()                                                 |
//|                                                                          |
//| Sets the current control back to normal, moves the cursor to the         |
//| previous control in the tab ring and sets that to focused.               |
//| Returns 0 if cursor is pointing to NULL, otherwise 1.                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XTabRing::Previous() {
	// Sanity Check:
	if (this->cursor == NULL)
		return(0);

	// Set focus of current control back to normal:
	this->cursor->control->SetNormal();

	// Move the cursor:
	this->cursor = this->cursor->prev;

	// 2. Sanity check:
	if (this->cursor == NULL)
		return(0);

	// Set the current control as focused:
	this->cursor->control->SetFocused();
	return(1);
} // end of int XTabRing::Previous()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XWinArgs::XWinArgs()                                                     |
//|                                                                          |
//| Initializes all the fields to some sensible values.                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
XWinArgs::XWinArgs() {
	this->display  = NULL;
	this->window   = 0;
	this->gc       = 0;
	this->colormap = 0;
	this->screen   = 0;
	this->xwindow  = NULL;
} // end of XWinArgs::XWinArgs()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XProperties::XProperties()                                               |
//|                                                                          |
//| Initializes all the fields to some sensible values.                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
XProperties::XProperties() {
	Reset();
} // end of XProperties::XProperties()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XProperties::Reset()                                                |
//|                                                                          |
//| Sets all the fields to some sensible values.                             |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XProperties::Reset() {
	this->x           = 0;
	this->y           = 0;
	this->width       = 0;
	this->height      = 0;
	this->edgesize    = 0;
	this->padding     = 0;
	this->taborder    = 0;
	this->itemspacing = 0;
	this->text = (char*)memset(this->textbuf, 0, sizeof(this->textbuf));
	this->font        = NULL;
	this->parent      = NULL;
} // end of void XProperties::Reset()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XProperties::SizeOfText()                                            |
//|                                                                          |
//| Returns the size of this->textbuf.                                       |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XProperties::SizeOfText() {
	return(sizeof(this->textbuf));
} // end of int XProperties::SizeOfText()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XObject::XObject()                                                       |
//|                                                                          |
//| Initializes the fields.                                                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
XObject::XObject() {
	Initialize();
} // end of XObject::XObject()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XObject::XObject(XWinArgs* xwinargs)                                     |
//|                                                                          |
//| Initializes the fields.                                                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
XObject::XObject(XWinArgs* winargs) {
	Initialize();
	this->display  = winargs->display;
	this->window   = winargs->window;
	this->gc       = winargs->gc;
	this->colormap = winargs->colormap;
	this->screen   = winargs->screen;
	this->xwindow  = winargs->xwindow;
} // end of XObject::XObject(XWinArgs* winargs)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XObject::Initialize()                                               |
//|                                                                          |
//| Initializes the basic XObject fields.                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XObject::Initialize() {
	// Initialize XObject's fields:
	memset(&this->error, 0, sizeof(this->error));
	memset(&this->uuid, 0, sizeof(this->uuid));
	this->type   = XW_TYPEOBJECT;
	this->x      = 0;
	this->y      = 0;
	this->width  = 0;
	this->height = 0;
	this->canvas = new XRectangle();

	// Generate a uuid:
	uuidgen(this->uuid);
} // end of void XObject::Initialize()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XObject::~XObject()                                                      |
//|                                                                          |
//| Frees the memory used by this->canvas.                                   |
//|                                                                          |
//+--------------------------------------------------------------------------+
XObject::~XObject() {
	XFree(this->canvas);
} // end of XObject::~XObject()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| char* XObject::Error()                                                   |
//|                                                                          |
//| Returns this->error.                                                     |
//|                                                                          |
//+--------------------------------------------------------------------------+
char* XObject::Error() {
	if (strcmp(this->error, "") == 0)
		return(NULL);
	else
		return((char*)&this->error);
} // end of char* XObject::Error()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XObject::Type()                                                      |
//|                                                                          |
//| Returns this->type.                                                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XObject::Type() {
	return(this->type);
} // end of int XObject::Type()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| char* XObject::Uuid()                                                    |
//|                                                                          |
//| Returns this->uuid.                                                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
char* XObject::Uuid() {
	return((char*)&this->uuid);
} // end of char* XObject::Uuid()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XObject::ClearError()                                               |
//|                                                                          |
//| Sets this->error to the empty string.                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XObject::ClearError() {
	memset(this->error, 0, sizeof(this->error));
} // end of void XObject::ClearError()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XObject::SetError(char* errormsg)                                   |
//|                                                                          |
//| Sets this->error to the error message passed in the argument.            |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XObject::SetError(char* errormsg) {
	strnzcpy(this->error, errormsg, sizeof(this->error));
} // end of void XObject::SetError(char* errormsg)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XCollection::XCollection()                                               |
//|                                                                          |
//| Instantiates a list object for the collectionof child controls.          |
//|                                                                          |
//+--------------------------------------------------------------------------+
XCollection::XCollection() {
	this->collection = new XList();
} // end of XCollection::XCollection()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XCollection::~XCollection()                                              |
//|                                                                          |
//| Frees the memory used by the collection.                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
XCollection::~XCollection() {
	this->collection->Clear();
	delete(this->collection);
} // end of XCollection::~XCollection()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XCollection::Add(XControl* control)                                  |
//|                                                                          |
//| Adds a control to the collection.                                        |
//| Returns 1 if the control could be added and 0 in case of errors.         |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XCollection::Add(XControl* control) {
	if (!this->collection->Add(control))
		return(0);
	return(1);
} // end of int XCollection::Add(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XCollection::Remove(XControl* control)                               |
//|                                                                          |
//| Removes a control from the collection.                                   |
//| Returns 1 if the control could be removed and 0 in case of errors.       |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XCollection::Remove(XControl* control) {
	if (!this->collection->Remove(control))
		return(0);
	return(1);
} // end of int XCollection::Remove(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XCollection::ChildCount()                                            |
//|                                                                          |
//| Returns this->collection->Count().                                       |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XCollection::ChildCount() {
	return(this->collection->Count());
} // end of int XCollection::ChildCount()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XWindow::XWindow()                                                       |
//|                                                                          |
//| The default constructor.                                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
XWindow::XWindow() {
} // end of XWindow::XWindow()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XWindow::XWindow(...) : XObject(), XCollection                           |
//|                                                                          |
//| Create the window and create and initialize videomemory.                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
XWindow::XWindow(int x, int y,
								 int width,
								 int height,
								 char* title,
								 int argc,
								 char** argv) : XObject(), XCollection() {
	// Initialize some fields:
	this->type            = XW_TYPEWINDOW;
	this->tabring         = new XTabRing();
	this->terminate       = 0;
	this->eventloopaction = 1;
	this->shiftdown       = 0;
	this->altdown         = 0;
	this->altgrdown       = 0;
	this->ctrldown        = 0;

	// Initialize fields from arguments:
	this->x              = x;
	this->y              = y;
	this->width          = width;
	this->height         = height;
	this->canvas->width  = width;
	this->canvas->height = height;
	
	// Open display:
	char* dspname = getenv("DISPLAY");
	if(!(this->display = XOpenDisplay(dspname))) {
		SetError("Can't open display.");
		return;
	}

	// Set the window's background color:
	this->bgcolor = WhitePixel(this->display, this->screen);

	// Create the window:
	if (!(this->window =
				XCreateSimpleWindow(this->display,
														RootWindow(this->display, this->screen),
														this->x, this->y, this->width, this->height, 4,
														BlackPixel(this->display, this->screen),
														this->bgcolor))) {
		SetError("Can't create window.");
		return;
  }

	// Get graphics context:
	XGCValues values;
	this->gc = XCreateGC(this->display, this->window, 0, &values);

	// Get the default colormap:
	this->colormap = DefaultColormap(display, DefaultScreen(display));

	// Create the background-storage pixmap and initialize it:
  std::cout << "create pixmap: ";
  std::cout << "XW_MAXWIDTH = " << XW_MAXWIDTH << " ";
  std::cout << "XW_MAXHEIGHT = " << XW_MAXHEIGHT << "\n";
	if (!(this->videomem = XCreatePixmap(this->display, this->window,
		XW_MAXWIDTH, XW_MAXHEIGHT, DefaultDepth(this->display, this->screen)))) {
		std::cerr << "Can't create pixmap.\n";
		SetError("Can't create pixmap.");
		return;
  }
  XSetFillStyle(this->display, this->gc, FillSolid);
 	XSetForeground(this->display, this->gc, this->bgcolor);
	XFillRectangle(this->display,
								 this->videomem,
								 this->gc,
								 0, 0, XW_MAXWIDTH, XW_MAXHEIGHT);

	// Create an icon:
	Pixmap icon = XCreatePixmap(this->display,
															this->window,
															48, 48,
															DefaultDepth(this->display, this->screen));
  XSetFillStyle(this->display, this->gc, FillSolid);
 	XSetForeground(this->display, this->gc,
								 WhitePixel(this->display, this->screen));
	XFillRectangle(this->display,
								 icon,
								 this->gc,
								 0, 0, 48, 48);
 	XSetForeground(this->display, this->gc,
								 BlackPixel(this->display, this->screen));
	XDrawRectangle(this->display,
								 icon,
								 this->gc,
								 5, 5, 38, 38);

	// Set window manager properties:
	if(!(sizehints = XAllocSizeHints())) {
		SetError("Can't allocate memory for sizehints.");
		return;
	}
	if(!(wmhints = XAllocWMHints())) {
		SetError("Can't allocate memory for wmhints.");
		return;
	}
	if(!(classhints = XAllocClassHint())) {
		SetError("Can't allocate memory for classhints.");
		return;
	}
	if(!XStringListToTextProperty(&title, 1, &windowname)) {
		SetError("Can't set windowname.");
		return;
	}
	sizehints->flags = PPosition | PSize | PMinSize;
	sizehints->x = this->x;
	sizehints->y = this->y;
	sizehints->width = this->width;
	sizehints->height = this->height;
	sizehints->min_width = XW_MINWIDTH;
	sizehints->min_height = XW_MINHEIGHT;
	wmhints->flags = IconPixmapHint;
	wmhints->icon_pixmap = icon;
	XSetWMProperties(this->display,
									 this->window,
									 &windowname,
									 &windowname,
									 argv,
									 argc,
									 sizehints,
									 wmhints,
									 classhints);

	// Set window manager protocols (needed to get close button to work):
	if ((wm_delete_window =
			 XInternAtom(this->display, "WM_DELETE_WINDOW", True)) == None) {
		SetError("Can't set WM_DELETE_WINDOW atom.");
		return;
	}
	XSetWMProtocols(this->display, this->window, &wm_delete_window, 1);

	// Set some more atoms needed to communicate with the controls:
	if ((wm_button_clicked =
			 XInternAtom(this->display, "WM_BUTTON_CLICKED", False)) == None) {
		SetError("Can't set WM_BUTTON_CLICKED atom.");
		return;
	}
	if ((wm_menu_clicked =
			 XInternAtom(this->display, "WM_MENU_CLICKED", False)) == None) {
		SetError("Can't set WM_MENU_CLICKED atom.");
		return;
	}
	if ((wm_text_changed =
			 XInternAtom(this->display, "WM_TEXT_CHANGED", False)) == None) {
		SetError("Can't set WM_TEXT_CHANGED atom.");
		return;
	}

	// Set the locale:
	if (setlocale(LC_ALL, "") == NULL) {
		SetError("Can't set locale.");
		return;
	}
	if (!XSupportsLocale()) {
		char locale[256];
		strnzcpy(locale, setlocale(LC_ALL, ""), sizeof(locale));
		char errmsg[512];
		memset(errmsg, 0, sizeof(errmsg));
		sprintf(errmsg, "Locale %s is not supported by X.", locale);
		SetError(errmsg);
		return;
	}
	if (XSetLocaleModifiers("") == NULL) {
		SetError("Can't set locale modifiers.");
		return;
	}

	// Create the font set:
	//char charsets[2048];
	//strnzcat(charsets, "-adobe-helvetica-*-r-*-*-*-120-*-*-*-*-*-*", sizeof(charsets));
	//strnzcat(charsets, ", -misc-fixed-*-r-*-*-*-130-*-*-*-*-*-*", sizeof(charsets));
	//strnzcat(charsets, "-*-*-*-R-Normal--*-180-100-100-*-*", sizeof(charsets));
	//char* charsets = "-adobe-helvetica-*-r-*-*-*-120-*-*-*-*-*-*";
	char* charsets = "-*-*-*-*-Normal-*-*-*-*-*-*-*-*-*";
	char* defaultset;
	char** missingsets;
	int missingsetscount = 0;
	this->fontset = XCreateFontSet(this->display, charsets, &missingsets,
		 &missingsetscount, &defaultset);
	if (this->fontset == NULL) {
		XFreeStringList(missingsets);
		std::cerr << "Failed to load charsets.";
		SetError("Failed to load charsets.");
		return;
	}
	if (missingsetscount > 0) {
		std::cerr << "Number of missing charsets: " << missingsetscount << "\n";
	}
	// Open an input method:
	XIM im;
	if ((im = XOpenIM(this->display, NULL, NULL, NULL)) == NULL) {
		SetError("Can't open input method.");
		return;
	}

	// Set style flags:
	XIMStyle appstyles;
	appstyles = XIMPreeditNone | XIMPreeditNothing | XIMPreeditArea;
  appstyles |= XIMStatusNone | XIMStatusNothing | XIMStatusArea;
	XIMStyles* imstyles;
	XGetIMValues(im, XNQueryInputStyle, &imstyles, NULL);
	XIMStyle style = 0;
	XIMStyle bestmatchedstyle = 0;
	for(int i = 0; i < imstyles->count_styles; ++i) {
		style = imstyles->supported_styles[i];
		if ((style & appstyles) == style)
			bestmatchedstyle = BestMatchedStyle(style, bestmatchedstyle);
	}
	if (bestmatchedstyle == 0) {
		SetError("Can't get a supported style.");
		return;
	}
	XFree(imstyles);
	fprintf(stderr, "___TRACE IN___\n");

	// Create the input context:
	XVaNestedList xvanestedlist =
		XVaCreateNestedList(0, XNFontSet, this->fontset, NULL);
	this->ic = XCreateIC(im,
                       XNInputStyle, bestmatchedstyle,
                       XNClientWindow, this->window,
                       XNPreeditAttributes, xvanestedlist,
                       XNStatusAttributes, xvanestedlist,
                       NULL);
	XFree(xvanestedlist);
	if (this->ic == NULL) {
		SetError("Can't create input context.");
		return;
	}
	XSetICFocus(this->ic);
	XFree(im);

	// Get the input event mask:
	long imeventmask;
	XGetICValues(this->ic, XNFilterEvents, &imeventmask, NULL);
 
	// Determine what input the window recognizes:
	XSelectInput(this->display, this->window,
							 ExposureMask |
							 KeyPressMask |
							 KeyReleaseMask |
							 ButtonPressMask |
							 ButtonReleaseMask |
							 Button1MotionMask |
							 Button2MotionMask |
							 Button3MotionMask |
							 StructureNotifyMask |
							 imeventmask);
} // end of XWindow::XWindow(...) : XObject(), XCollection()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XWindow::~XWindow()                                                      |
//|                                                                          |
//| Deletes all controls (frees the memory used by them), then frees all x   |
//| resources used and closes the window.                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
XWindow::~XWindow() {
	// Delete the tab ring:
	this->tabring->Clear();
	delete(tabring);

	// Free x resources and close window:
	XDeleteProperty(this->display, this->window, this->wm_delete_window);
	XDeleteProperty(this->display, this->window, this->wm_button_clicked);
	XFreePixmap(this->display, this->videomem);
	XFree(this->sizehints);
	XFree(this->wmhints);
	XFree(this->classhints);
	XFreeGC(this->display, this->gc);
	XCloseDisplay(this->display);
} // end of XWindow::~XWindow()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XWindow::Run()                                                       |
//|                                                                          |
//| Implements the event handler. Everything happens here.                   |
//| Returns 1 if the event loop was terminated by setting this->terminate    |
//| to 1 and 0 if the event loop was simply left because of errors.          |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XWindow::Run() {
	// Sanity check:
	if (Error())
		return(0);

	// Map window:
 	XMapWindow(this->display, this->window);

	// Start the event loop:
	XEvent event;
	while (!this->terminate) {
		// Do something besides waiting for events:
		if (this->eventloopaction)
			EventLoopAction();

		// Check for a new event:
		XCheckIfEvent(this->display, &event, &pred, 0);

		// Test if one of the controls acts on this event:
		this->collection->Reset();
		while (this->collection->Item() != NULL) {
			if (this->collection->Item()->EventHandler(&event)) {
				event.type = 0;
				break;
			}
			this->collection->Next();
		}
		if (event.type == 0)
			continue;

		// No control acted on this event -> let the window act on it:
		char keystr[XW_TEXTSIZESYSTEMMSG];
		memset(&keystr, 0, sizeof(keystr));
		KeySym keysym;
		switch (event.type) {
		case ClientMessage:
			if (event.xclient.data.l[0] == (int)wm_delete_window) {
				Terminate();
				event.type = 0;
				break;
			}
			EvHandlerClientMessage((XClientMessageEvent*)&event.xclient);
			event.type = 0;
			break;
		case Expose:
			while (XCheckTypedEvent(this->display, Expose, &event));
			event.type = 0;
			// Reset any menu dropdowns:
			this->collection->Reset();
			while(this->collection->Item() != NULL) {
				if (this->collection->Item()->Type() == XW_TYPEMENUBAR) {
					((XMenuBar*)this->collection->Item())->SetActiveMenuTo(NULL);
					break;
				}
			}
			// Repaint rest of window:
			Paint();
			break;
		case ConfigureNotify:
			// Mark event as acted upon:
			event.type = 0;

			// If the window's size has not changed do nothing:
			if ((event.xconfigure.width == this->width) ||
					(event.xconfigure.height == this->height))
				break;

			// Set window to new size and repaint it:
			this->width = event.xconfigure.width;
			this->height = event.xconfigure.height;
			if (this->width > XW_MAXWIDTH)
				this->width = XW_MAXWIDTH;
			if (this->height > XW_MAXHEIGHT)
				this->height = XW_MAXHEIGHT;

			// Redraw menubar and statusbar:
			this->collection->Reset();
			while (this->collection->Item() != NULL) {
				if ((this->collection->Item()->Type() == XW_TYPEMENUBAR) ||
						(this->collection->Item()->Type() == XW_TYPESTATUSBAR)) {
					this->collection->Item()->Paint();
				}
				this->collection->Next();
			}
			break;
		case ButtonPress:
			if (event.xbutton.button == Button1) {
				EvHandlerButton1Press((XButtonEvent*)&event.xbutton);
			}
			if (event.xbutton.button == Button2) {
				EvHandlerButton2Press((XButtonEvent*)&event.xbutton);
			}
			if (event.xbutton.button == Button3) {
				EvHandlerButton3Press((XButtonEvent*)&event.xbutton);
			}
			event.type = 0;
			break;
		case ButtonRelease:
			if (event.xbutton.button == Button1) {
				// If a dropdown is visible close it:
				this->collection->Reset();
				while(this->collection->Item() != NULL) {
					if (this->collection->Item()->Type() == XW_TYPEMENUBAR) {
						XMenuBar* obj = (XMenuBar*)this->collection->Item();
						if (obj->ActiveMenu() != NULL) {
							obj->SetActiveMenuTo(NULL);
							event.type = 0;
							break;
						}
					}
					this->collection->Next();
				}
				if (event.type == 0)
					break;
				EvHandlerButton1Release((XButtonEvent*)&event.xbutton);
			}
			if (event.xbutton.button == Button2) {
				EvHandlerButton2Release((XButtonEvent*)&event.xbutton);
			}
			if (event.xbutton.button == Button3) {
				EvHandlerButton3Release((XButtonEvent*)&event.xbutton);
			}
			event.type = 0;
			break;
		case MotionNotify:
			if (event.xbutton.state == Button1Mask) {
				EvHandlerButton1MotionNotify((XButtonEvent*)&event.xbutton);
			}
			if (event.xbutton.state == Button2Mask) {
				EvHandlerButton2MotionNotify((XButtonEvent*)&event.xbutton);
			}
			if (event.xbutton.state == Button3Mask) {
				EvHandlerButton3MotionNotify((XButtonEvent*)&event.xbutton);
			}
			event.type = 0;
			break;
		case KeyPress:
			keysym = XKeycodeToKeysym(this->display, event.xkey.keycode, 0);
			strnzcpy(keystr, XKeysymToString(keysym), sizeof(keystr));
			// Set the system key fields:
			if ((strcmp(keystr, "Shift_L") == 0) ||
					(strcmp(keystr, "Shift_R") == 0))
				this->shiftdown = 1;
			if ((strcmp(keystr, "Alt_L") == 0) ||
					(strcmp(keystr, "Alt_R") == 0))
				this->altdown = 1;
			if (strcmp(keystr, "Mode_switch") == 0)
				this->altgrdown = 1;
			if ((strcmp(keystr, "Control_L") == 0) ||
					(strcmp(keystr, "Control_R") == 0))
				this->ctrldown = 1;
			// Tab is used to access controls by setting focus on them:
			if (strcmp(keystr, "Tab") == 0) {
				if ((!this->altdown) && (!this->altgrdown) && (!this->ctrldown)) {
					if (this->shiftdown)
						this->tabring->Previous();
					else
						this->tabring->Next();
					event.type = 0;
					break;
				}
			}
			// Some other key:
			EvHandlerKeyPressed((XKeyEvent*)&event.xkey);
			event.type = 0;
			break;
		case KeyRelease:
			keysym = XKeycodeToKeysym(this->display, event.xkey.keycode, 0);
			strnzcpy(keystr, XKeysymToString(keysym), sizeof(keystr));
			// Reset the system key fields:
			if ((strcmp(keystr, "Shift_L") == 0) ||
					(strcmp(keystr, "Shift_R") == 0))
				this->shiftdown = 0;
			if ((strcmp(keystr, "Alt_L") == 0) ||
					(strcmp(keystr, "Alt_R") == 0))
				this->altdown = 0;
			if (strcmp(keystr, "Mode_switch") == 0)
				this->altgrdown = 0;
			if ((strcmp(keystr, "Control_L") == 0) ||
					(strcmp(keystr, "Control_R") == 0))
				this->ctrldown = 0;
			// Some other key:
			EvHandlerKeyReleased((XKeyEvent*)&event.xkey);
			event.type = 0;
			break;
		default:
			event.type = 0;
			break;
		} // end switch(event.type)
	} // while(!this->terminate)

	// Unmap window:
 	XUnmapWindow(this->display, this->window);
	return(1);
} // end of int XWindow::Run()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XWindow::Add(XControl* control)                                      |
//|                                                                          |
//| Adds a new control to the collection and to the tab ring.                |
//| Returns 1 if the control could be added and 0 in case of errors.         |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XWindow::Add(XControl* control) {
	// Adding is done in the base class:
	if (!XCollection::Add(control))
		return(0);

	// Some controls are not accepted in the tabring:
	if ((control->Type() == XW_TYPESTATUSBAR) ||
			(control->Type() == XW_TYPESTATUSBARITEM))
		return(1);

	// Add the control to the tab ring:
	if (!this->tabring->Insert(control)) {
		SetError("Could not add control to tab ring.");
		return(0);
	}
	return(1);
} // end of int XWindow::Add(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XWindow::Remove(XControl* control)                                   |
//|                                                                          |
//| Removes a control from the collection and from the tab ring.             |
//| Returns 1 if the control could be removed and 0 in case of errors.       |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XWindow::Remove(XControl* control) {
	// Removing is done in the base class:
	if (!XCollection::Remove(control))
		return(0);

	// Remove the control from the tab ring:
	if (!this->tabring->Remove(control)) {
		SetError("Could not add control to tab ring.");
		return(0);
	}
	return(1);
} // end of int XWindow::Remove(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::SetClipRegionToCanvas()                                    |
//|                                                                          |
//| Sets the clipping region to the rectangle defined by this->canvas. This  |
//| is the area that's acceptable for drawing (controls can "isolate" space  |
//| for themselves with the ShrinkCanvas() method - statusbar is an example  |
//| for this).                                                               |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::SetClipRegionToCanvas() {
	XSetClipRectangles(this->display,this->gc,0,0,this->canvas,1,Unsorted);
} // end of void XWindow::SetClipRegionToCanvas()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::SetClipRegionToWindow()                                    |
//|                                                                          |
//| Sets the clipping region to the size of the window.                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::SetClipRegionToWindow() {
	XRectangle* winrect = new XRectangle();
 	winrect->x = 0;
 	winrect->y = 0;
 	winrect->width  = this->width;
 	winrect->height = this->height;
	XSetClipRectangles(this->display,this->gc, 0, 0, winrect, 1, Unsorted);
	XFree(winrect);
} // end of void XWindow::SetClipRegionToWindow()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::UnsetClipRegion()                                          |
//|                                                                          |
//| Sets the clipping region to the maximum size the window can have.        |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::UnsetClipRegion() {
	XRectangle* winrect = new XRectangle();
 	winrect->x = 0;
 	winrect->y = 0;
 	winrect->width  = XW_MAXWIDTH;
 	winrect->height = XW_MAXHEIGHT;
	XSetClipRectangles(this->display,this->gc, 0, 0, winrect, 1, Unsorted);
	XFree(winrect);
} // end of void XWindow::UnsetClipRegion()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XWindow::SetFocusTo(XControl* control)                               |
//|                                                                          |
//| Sets the focus to the control specified in the argument. Any other       |
//| control that is currently focused is set to normal.                      |
//| Returns 1 if the focus could be set seuccesfuly and 0 in case of errors. |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XWindow::SetFocusTo(XControl* control) {
	return(this->tabring->SetFocusTo(control));
} // end of int XWindow::SetFocusTo(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XWindow::ShrinkCanvas(int side, int shrink)                          |
//|                                                                          |
//| Reduces the canvas on the given side by shrink pixels. The clipping      |
//| region is automatically set to this new canvas as well.                  |
//| Controls (e.g. the statusbar) can use this method to "isolate" space     |
//| in the window for themselves into which the other controls then can't    |
//| draw anymore.                                                            |
//| Returns 1 if this->canvas could be set succesfully and 0 if side could   |
//| not be resolved.                                                         |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XWindow::ShrinkCanvas(int side, int shrink) {
	switch (side) {
	case XW_CANVASLEFT:
		this->canvas->x += shrink;
		this->canvas->width -= shrink;
		break;
	case XW_CANVASRIGHT:
		this->canvas->width -= shrink;
		break;
	case XW_CANVASTOP:
		this->canvas->y += shrink;
		this->canvas->height -= shrink;
		break;
	case XW_CANVASBOTTOM:
		this->canvas->height -= shrink;
		break;
	default:
		return(0);
		break;
	}
	XSetClipRectangles(this->display,this->gc,0,0,this->canvas,1,Unsorted);
	return(1);
} // end of int XWindow::ShrinkCanvas(int side, int shrink)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XWindow::ExpandCanvas(int side, int expand)                          |
//|                                                                          |
//| Expands the canvas on the given side by shrink pixels. The clipping      |
//| region is automatically set to this new canvas as well.                  |
//| Controls (e.g. the statusbar) should use this method in their destructor |
//| if they also used ShrinkCanvas().                                        |
//| Returns 1 if this->canvas could be set succesfully and 0 if side could   |
//| not be resolved.                                                         |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XWindow::ExpandCanvas(int side, int expand) {
	switch (side) {
	case XW_CANVASLEFT:
		this->canvas->x -= expand;
		this->canvas->width += expand;
		break;
	case XW_CANVASRIGHT:
		this->canvas->width += expand;
		break;
	case XW_CANVASTOP:
		this->canvas->y -= expand;
		this->canvas->height += expand;
		break;
	case XW_CANVASBOTTOM:
		this->canvas->height += expand;
		break;
	default:
		return(0);
		break;
	}
	XSetClipRectangles(this->display,this->gc,0,0,this->canvas,1,Unsorted);
	return(1);
} // end of int XWindow::ExpandCanvas(int side, int expand)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XWindow::Width()                                                     |
//|                                                                          |
//| Returns this->width.                                                     |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XWindow::Width() {
	return(this->width);
} // end of int XWindow::Width()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XWindow::Height()                                                    |
//|                                                                          |
//| Returns this->height.                                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XWindow::Height() {
	return(this->height);
} // end of int XWindow::Height()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| unsigned long XWindow::BGColor()                                         |
//|                                                                          |
//| Returns this->bgcolor.                                                   |
//|                                                                          |
//+--------------------------------------------------------------------------+
unsigned long XWindow::BGColor() {
	return(this->bgcolor);
} // end of unsigned long XWindow::BGColor()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::Paint()                                                    |
//|                                                                          |
//| Redraws the window and all the controls associated with it.              |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::Paint() {
	// Draw the window:
	XClearWindow(this->display, this->window);
	XCopyArea(this->display, this->videomem, this->window, this->gc,
						0, 0, this->width, this->height, 0, 0);
} // end of void XWindow::Paint()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::Terminate()                                                |
//|                                                                          |
//| Sets this->terminate to 1, thus stopping execution of the event loop.    |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::Terminate() {
	this->terminate = 1;
} // end void XWindow::Terminate()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EventLoopAction()                                          |
//|                                                                          |
//| Whatever happens in this method is executed in every cycle of the event  |
//| loop.                                                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EventLoopAction() {
} // end void XWindow::EventLoopAction()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EvHandlerClientMessage(XClientMessageEvent* event)         |
//|                                                                          |
//| The event handler for events triggered by the controls.                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EvHandlerClientMessage(XClientMessageEvent* event) {
  fprintf(stdout, "in XWindow::EvHandlerClientMessage\n");
} // end of void XWindow::EvHandlerClientMessage(XClientMessageEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EvHandlerButton1Press(XButtonEvent* event)                 |
//|                                                                          |
//| The event handler for "left mouse button pressed" events.                |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EvHandlerButton1Press(XButtonEvent* event) {
  fprintf(stdout, "in XWindow::EvHandlerButton1Press\n");
} // end of void XWindow::EvHandlerButton1Press(XButtonEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EvHandlerButton2Press(XButtonEvent* event)                 |
//|                                                                          |
//| The event handler for "middle mouse button pressed" events.              |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EvHandlerButton2Press(XButtonEvent* event) {
} // end of void XWindow::EvHandlerButton2Press(XButtonEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EvHandlerButton3Press(XButtonEvent* event)                 |
//|                                                                          |
//| The event handler for "right mouse button pressed" events.               |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EvHandlerButton3Press(XButtonEvent* event) {
} // end of void XWindow::EvHandlerButton3Press(XButtonEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EvHandlerButton1Release(XButtonEvent* event)               |
//|                                                                          |
//| The event handler for "left mouse button up" events.                     |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EvHandlerButton1Release(XButtonEvent* event) {
} // end of void XWindow::EvHandlerButton1Release(XButtonEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EvHandlerButton2Release(XButtonEvent* event)               |
//|                                                                          |
//| The event handler for "middle mouse button up" events.                   |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EvHandlerButton2Release(XButtonEvent* event) {
} // end of void XWindow::EvHandlerButton2Release(XButtonEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EvHandlerButton3Release(XButtonEvent* event)               |
//|                                                                          |
//| The event handler for "right mouse button up" events.                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EvHandlerButton3Release(XButtonEvent* event) {
} // end of void XWindow::EvHandlerButton3Release(XButtonEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EvHandlerButton1MotionNotify(XButtonEvent* event)          |
//|                                                                          |
//| The event handler for "left mouse button down and moved" events.         |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EvHandlerButton1MotionNotify(XButtonEvent* event) {
} // end of void XWindow::EvHandlerButton1MotionNotify(XButtonEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EvHandlerButton2MotionNotify(XButtonEvent* event)          |
//|                                                                          |
//| The event handler for "middle mouse button down and moved" events.       |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EvHandlerButton2MotionNotify(XButtonEvent* event) {
} // end of void XWindow::EvHandlerButton2MotionNotify(XButtonEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EvHandlerButton3MotionNotify(XButtonEvent* event)          |
//|                                                                          |
//| The event handler for "right mouse button down and moved" events.        |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EvHandlerButton3MotionNotify(XButtonEvent* event) {
} // end of void XWindow::EvHandlerButton3MotionNotify(XButtonEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EvHandlerKeyPressed(XKeyEvent* event)                      |
//|                                                                          |
//| The event handler for "key down" events.                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EvHandlerKeyPressed(XKeyEvent* event) {
} // end of void XWindow::EvHandlerKeyPressed(XKeyEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XWindow::EvHandlerKeyReleased(XKeyEvent* event)                     |
//|                                                                          |
//| The event handler for "key up" events.                                   |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XWindow::EvHandlerKeyReleased(XKeyEvent* event) {
} // end of void XWindow::EvHandlerKeyReleased(XKeyEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XIMStyle XWindow::BestMatchedStyle(XIMStyle style1, XIMStyle style2)     |
//|                                                                          |
//| Selectes the "better" one of the two styles.                             |
//|                                                                          |
//+--------------------------------------------------------------------------+
XIMStyle XWindow::BestMatchedStyle(XIMStyle style1, XIMStyle style2) {
	// Simple solution possible?
	if (style1 == 0) return style2;
	if (style2 == 0) return style1;

	// If styles are equal...
	XIMStyle preedit = XIMPreeditArea |
		XIMPreeditCallbacks |
		XIMPreeditPosition |
		XIMPreeditNothing |
		XIMPreeditNone;
	XIMStyle status = XIMStatusArea |
		XIMStatusCallbacks |
		XIMStatusNothing |
		XIMStatusNone;
	if ((style1 & (preedit | status)) == (style2 & (preedit | status)))
		return(style1);


	XIMStyle s1 = style1 & preedit;
  XIMStyle s2 = style2 & preedit;
	if (s1 != s2) {
		// Preedit flags are different => compare on preedit flags:
		if (s1 | s2 | XIMPreeditCallbacks) {
			if (s1 == XIMPreeditCallbacks)
				return(style1);
			else
				return(style2);
		}
    else if (s1 | s2 | XIMPreeditPosition) {
			if (s1 == XIMPreeditPosition)
				return(style1);
			else
				return(style2);			
		}
    else if (s1 | s2 | XIMPreeditArea) {
			if (s1 == XIMPreeditArea)
				return(style1);
			else
				return(style2);			
		}
    else if (s1 | s2 | XIMPreeditNothing) {
			if (s1 == XIMPreeditNothing)
				return(style1);
			else
				return(style2);			
		}
	} // end if (s1 != s2)
	else {
		// Preedit flags are the same => compare on status flags:
		s1 = style1 & status;
		s2 = style2 & status;
		if (s1 | s2 | XIMStatusCallbacks) {
			if (s1 == XIMStatusCallbacks)
				return(style1);
			else
				return(style2);
		}
		else if (s1 | s2 | XIMStatusArea) {
			if (s1 == XIMStatusArea)
				return(style1);
			else
				return(style2);
		}
		else if (s1 | s2 | XIMStatusNothing) {
			if (s1 == XIMStatusNothing)
				return(style1);
			else
				return(style2);
		}
	}
	// Just in case:
	return(style1);
} // end of XIMStyle XWindow::BestMatchedStyle(XIMStyle style1, style2)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XAppRoot::XAppRoot()                                                     |
//|                                                                          |
//| Seeds the rand() function whenever a new application is instantiated.    |
//|                                                                          |
//+--------------------------------------------------------------------------+
XAppRoot::XAppRoot() {
	srand(time(NULL));
} // end of XAppRoot::XAppRoot()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XApplication::XApplication()                                             |
//|                                                                          |
//| The default constructor.                                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
XApplication::XApplication() {
} // end of XApplication::XApplication()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XApplication::XApplication(...) : XAppRoot(), XWindow(...)               |
//|                                                                          |
//| The standard constructor just initializes some fields.                   |
//|                                                                          |
//+--------------------------------------------------------------------------+
XApplication::XApplication(int x, int y,
													 int width,
													 int height,
													 char* title,
													 int argc, char** argv) :
	XAppRoot(), XWindow(x, y, width, height, title, argc, argv) {
	// Set type and progname:
	this->type = XW_TYPEAPPLICATION;
	strnzcpy(this->progname, argv[0], sizeof(this->progname));
} // end of XApplication::XApplication(...) : XAppRoot(), XWindow(...)


//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XApplication::Run()                                                  |
//|                                                                          |
//| Adds error handling to the inherited Run() method.                       |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XApplication::Run() {
	if (!XWindow::Run()) {
		fprintf(stderr, "%s: %s\n", this->progname, this->error);
		return(0);
	}
	return(1);
} // end of int XApplication::Run()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XControl::XControl()                                                     |
//|                                                                          |
//| The default constructor.                                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
XControl::XControl() {
} // end of XControl::XControl()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XControl::XControl(...) : XObject()                                      |
//|                                                                          |
//| The constructor for standard controls like buttons etc.                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
XControl::XControl(XWinArgs* winargs, XProperties* properties) :
	XObject(winargs) {
	this->type        = XW_TYPECONTROL;
	this->framesize   = XW_FRAMESIZENORMAL;
	this->focused     = 0;
	this->x           = properties->x;
	this->y           = properties->y;
	this->width       = properties->width;
	this->height      = properties->height;
	this->edgesize    = properties->edgesize;
	this->padding     = properties->padding;
	this->taborder    = properties->taborder;
	this->itemspacing = properties->itemspacing;
	this->font        = properties->font;
	strnzcpy(this->text, properties->text, sizeof(this->text));
	this->parent      = (XObject*)properties->parent;

	// Allocate the colors:
	AllocateColors();
} // end of XControl::XControl(...) : XObject()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XControl::~XControl()                                                    |
//|                                                                          |
//| Frees any x resources allocated in the constructor.                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
XControl::~XControl() {
	XFreeColors(this->display, this->colormap, &framecolor, 1, 0);
	XFreeColors(this->display, this->colormap, &edgecolorlight, 1, 0);
	XFreeColors(this->display, this->colormap, &edgecolordark, 1, 0);
	XFreeColors(this->display, this->colormap, &backgroundcolor, 1, 0);
	XFreeColors(this->display, this->colormap, &foregroundcolor, 1, 0);
	XFreeColors(this->display, this->colormap, &disabledcolor, 1, 0);
} // end of XControl::~XControl()


//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XControl::Paint()                                                   |
//|                                                                          |
//| The paint methods draws the control in it's current state.               |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XControl::Paint() {
} // end of void XControl::Paint()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XControl::EventHandler(XEvent* event)                                |
//|                                                                          |
//| The event handler method checks if the event happened "inside" the       |
//| control and acts accordingly if it did.                                  |
//| It should return 1 if the event happened inside this control's area      |
//| _and_ was processed by it, zero if not.                                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XControl::EventHandler(XEvent* event) {
	return(0);
} // end of int XControl::EventHandler(XEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XObject* XControl::Parent()                                              |
//|                                                                          |
//| Returns this->parent.                                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
XObject* XControl::Parent() {
	return(this->parent);
} // end of XObject* XControl::Parent()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| char* XControl::Text()                                                   |
//|                                                                          |
//| Returns this->text.                                                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
char* XControl::Text() {
	if (strcmp(this->text, "") == 0)
		return(NULL);
	else
		return((char*)&this->text);
} // end of char* XControl::Text()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XControl::SetText(char* newtext)                                    |
//|                                                                          |
//| If newtext is different from this->text, this->text is set to newtext    |
//| and the control is repainted.                                            |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XControl::SetText(char* newtext) {
	if (strcmp(newtext, this->text) == 0)
		return;
	strnzcpy(this->text, newtext, sizeof(this->text));
	Paint();
} // end of void XControl::SetText(char* newtext)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XControl::SetFocused()                                              |
//|                                                                          |
//| Sets this->focused to 1. If this changes the control's state the control |
//| is repainted.                                                            | 
//|                                                                          |
//+--------------------------------------------------------------------------+
void XControl::SetFocused() {
	if (this->focused)
		return;
	this->focused = 1;
	this->framesize = XW_FRAMESIZEFOCUSED;
	Paint();
} // end of void XControl::SetFocused()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XControl::SetNormal()                                               |
//|                                                                          |
//| Sets this->focused to 0. If this changes the control's state the control |
//| is repainted.                                                            |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XControl::SetNormal() {
	if (!this->focused)
		return;
	this->focused = 0;
	this->framesize = XW_FRAMESIZENORMAL;
	Paint();
} // end of void XControl::SetNormal()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XControl::TextWidth()                                                |
//|                                                                          |
//| Returns the width of this->text.                                         |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XControl::TextWidth() {
	if (strcmp(this->text, "") == 0)
		return(0);
	return(XTextWidth(this->font, this->text, strlen(this->text)));
} // end of int XControl::TextWidth()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XControl::TextHeight()                                               |
//|                                                                          |
//| Returns the height of this->text.                                        |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XControl::TextHeight() {
	if (this->font == NULL)
		return(0);
	return(this->font->ascent + this->font->descent);
} // end of int XControl::TextHeight()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XControl::TabOrder()                                                 |
//|                                                                          |
//| Returns this->taborder.                                                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XControl::TabOrder() {
	return(this->taborder);
} // end of int XControl::TabOrder()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XControl::Focused()                                                  |
//|                                                                          |
//| Returns this->focused.                                                   |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XControl::Focused() {
	return(this->focused);
} // end of int XControl::Focused()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XControl::Padding()                                                  |
//|                                                                          |
//| Returns this->padding.                                                   |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XControl::Padding() {
	return(this->padding);
} // end of int XControl::Padding()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XControl::Edgesize()                                                 |
//|                                                                          |
//| Returns this->edgesize.                                                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XControl::Edgesize() {
	return(this->edgesize);
} // end of int XControl::Edgesize()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XControl::AllocateColors()                                          |
//|                                                                          |
//| Allocates the system colors of this control.                             |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XControl::AllocateColors() {
	XColor rgb;
	rgb.red = 0x0000; rgb.green = 0x0000; rgb.blue = 0x0000;
	XAllocColor(this->display, this->colormap, &rgb);
 	this->framecolor = rgb.pixel;
	rgb.red = 0xf000; rgb.green = 0xf000; rgb.blue = 0xf000;
	XAllocColor(this->display, this->colormap, &rgb);
 	this->edgecolorlight = rgb.pixel;
	rgb.red = 0x8000; rgb.green = 0x8000; rgb.blue = 0x8000;
	XAllocColor(this->display, this->colormap, &rgb);
 	this->edgecolordark = rgb.pixel;
	rgb.red = 0xc000; rgb.green = 0xc000; rgb.blue = 0xc000;
	XAllocColor(this->display, this->colormap, &rgb);
 	this->backgroundcolor = rgb.pixel;
	rgb.red = 0x0000; rgb.green = 0x0000; rgb.blue = 0x0000;
	XAllocColor(this->display, this->colormap, &rgb);
 	this->foregroundcolor = rgb.pixel;
	rgb.red = 0x8000; rgb.green = 0x8000; rgb.blue = 0x8000;
	XAllocColor(this->display, this->colormap, &rgb);
 	this->disabledcolor = rgb.pixel;
} // end of void XControl::AllocatesColors()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XControl::SetClipRegionToObject()                                   |
//|                                                                          |
//| Sets the clipping region to the rectangle defined by this->x, this->y,   |
//| this->width and this->height. This is the maximum area any control needs |
//| to draw itself.                                                          |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XControl::SetClipRegionToObject() {
	XRectangle* rect = new XRectangle();
	rect->x      = this->x;
	rect->y      = this->y;
	rect->width  = this->width;
	rect->height = this->height;
	XSetClipRectangles(this->display, this->gc, 0, 0, rect, 1, Unsorted);
	XFree(rect);
} // end of void XControl::SetClipRegionToObject()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XControl::SetClipRegionToCanvas()                                   |
//|                                                                          |
//| Sets the clipping region to the rectangle defined by this->canvas. This  |
//| is the area inside any edges which can be used for drawing.              |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XControl::SetClipRegionToCanvas() {
	XSetClipRectangles(this->display,this->gc,0,0,this->canvas,1,Unsorted);
} // end of void XControl::SetClipRegionToCanvas()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XButton::XButton()                                                       |
//|                                                                          |
//| The default constructor.                                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
XButton::XButton() {
} // end of XButton::XButton()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XButton::XButton(...) : XControl(...)                                    |
//|                                                                          |
//| The button's fields are initialized and the button is drawn.             |
//|                                                                          |
//+--------------------------------------------------------------------------+
XButton::XButton(XWinArgs* winargs, XProperties* properties) :
	XControl(winargs, properties) {
	// Initialize fields:
	this->type     = XW_TYPEBUTTON;
	this->pressed  = 0;
	this->disabled = 0;

	// Set the atom needed to communicate with the parent:
	if ((wm_button_clicked =
			 XInternAtom(this->display, "WM_BUTTON_CLICKED", True)) == None) {
		SetError("can't set WM_BUTTON_CLICKED atom.");
		return;
	}

	// Add to the list of controls kept by the parent:
	if (!((XWindow*)this->parent)->Add(this)) {
		SetError("Button could not be added to parent collection.");
		return;
	}

	// Draw the button:
	this->Paint();
} // end of XButton::XButton(...) : XControl(...)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XButton::Paint()                                                    |
//|                                                                          |
//| Paints the button depending on it's current state.                       |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XButton::Paint() {
	// Determine canvas area:
	this->canvas->x = this->x + this->framesize + this->edgesize + this->padding;
	this->canvas->y = this->y + this->framesize + this->edgesize + this->padding;
	this->canvas->width   = this->width - (2 * this->framesize);
	this->canvas->width  -= (2 * this->edgesize);
	this->canvas->width  -= (2 * this->padding);
	this->canvas->height  = this->height - (2 * this->framesize);
	this->canvas->height -= (2 * this->edgesize);
	this->canvas->height -= (2 * this->padding);

	// Draw outer frame:
  XSetLineAttributes(this->display, this->gc, 1,
										 LineSolid, CapNotLast, JoinMiter);
 	XSetForeground(this->display, this->gc, this->framecolor);
	for(int i = 0; i < this->framesize; ++i) {
		XDrawRectangle(this->display,
									 this->xwindow->videomem,
									 this->gc,
									 this->x+i, this->y+i,
									 this->width-(2*i), this->height-(2*i));
	}

	// Draw button background:
  XSetFillStyle(this->display, this->gc, FillSolid);
 	XSetForeground(this->display, this->gc, this->backgroundcolor);
	XFillRectangle(this->display,
								 this->xwindow->videomem,
								 this->gc,
								 this->x+this->framesize, this->y+this->framesize,
								 this->width-(2*this->framesize),
								 this->height-(2*this->framesize));

	// Draw edges:
	for (int i = 0; i < this->edgesize; ++i) {
		if (this->pressed)
			XSetForeground(this->display, this->gc, this->edgecolordark);
		else
			XSetForeground(this->display, this->gc, this->edgecolorlight);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+this->framesize,             this->y+this->framesize+i,
							this->x+this->width-this->framesize, this->y+this->framesize+i);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+this->framesize+i, this->y+this->framesize,
							this->x+this->framesize+i, this->y+this->height-this->framesize);
		if (this->pressed)
			XSetForeground(this->display, this->gc, this->edgecolorlight);
		else
			XSetForeground(this->display, this->gc, this->edgecolordark);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+(2*this->framesize)+i,
							this->y+this->height-this->framesize-i,
							this->x+this->width-this->framesize,
							this->y+this->height-this->framesize-i);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+this->width-this->framesize-i,
							this->y+(2*this->framesize)+i,
							this->x+this->width-this->framesize-i,
							this->y+this->height-this->framesize);
	}

	// Draw text:
	int textheight = this->font->ascent + this->font->descent;
	int textwidth = XTextWidth(this->font, this->text, strlen(this->text));
	int internalwidth =
		this->width - (2*this->framesize) - (2*this->edgesize);
	int internalheight =
		this->height - (2*this->framesize) - (2*this->edgesize);
	int tx = this->x + this->framesize + this->edgesize;
	tx += (internalwidth - textwidth)/2;
	if (this->pressed)
		tx += this->edgesize;
	int ty =
		this->y + this->framesize + this->edgesize + this->font->ascent;
	ty += (internalheight - textheight)/2;
	if (this->pressed)
		ty += this->edgesize;
	if (this->disabled)
		XSetForeground(this->display, this->gc, this->disabledcolor);
	else
		XSetForeground(this->display, this->gc, this->foregroundcolor);
	XSetFont(this->display, this->gc, this->font->fid);
	SetClipRegionToCanvas();
	XDrawString(this->display, this->xwindow->videomem, this->gc,
							tx, ty, this->text, strlen(this->text));
	if (this->focused) {
		XSetLineAttributes(this->display, this->gc, 1,
											 LineOnOffDash, CapNotLast, JoinMiter);
		XDrawRectangle(this->display, this->xwindow->videomem, this->gc,
									 tx-1, ty-this->font->ascent-1,
									 textwidth+1, textheight+1);
	}
	this->xwindow->SetClipRegionToWindow();

	// Finally copy the button back onto main window:
	XCopyArea(this->display, this->xwindow->videomem, this->window, this->gc,
						this->x, this->y, this->width, this->height, this->x, this->y);
} // end of void XButton::Paint()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XButton::EventHandler(XEvent* event)                                 |
//|                                                                          |
//| The event handler for the button.                                        |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XButton::EventHandler(XEvent* event) {
	// If the button is disabled, no events are recognized:
	int processedevent = 0;
	if (this->disabled)
		return(processedevent);

	// The button is not disabled -> process events:
	char keystr[XW_TEXTSIZESYSTEMMSG];
	memset(&keystr, 0, sizeof(keystr));
	KeySym keysym;
	XEvent xevent;
	memset(&xevent, 0, sizeof(xevent));
	xevent.xclient.type = ClientMessage;
	xevent.xclient.window = this->window;
	xevent.xclient.message_type = this->wm_button_clicked;
	xevent.xclient.format = 32;
	strcpy(xevent.xclient.data.b, this->uuid);
	xevent.xclient.data.l[0] = this->wm_button_clicked;
	xevent.xclient.data.l[1] = (long)this;
	switch (event->type) {
	case ButtonPress:
		if ((event->xbutton.x < this->x) ||
				(event->xbutton.y < this->y) ||
				(event->xbutton.x > this->x+this->width) ||	
				(event->xbutton.y > this->y+this->height))
			break;
		if (event->xbutton.button == Button1) {
			this->pressed = 1;
			if (!this->focused)
				this->xwindow->SetFocusTo(this);
			else
				Paint();
			processedevent = 1;
		}
		break;
	case ButtonRelease:
		if ((event->xbutton.x < this->x) ||
				(event->xbutton.y < this->y) ||
				(event->xbutton.x > this->x+this->width) ||	
				(event->xbutton.y > this->y+this->height))
			break;
		if (event->xbutton.button == Button1) {
			if (this->pressed) {
				this->pressed = 0;
				Paint();
				XSendEvent(this->display, this->window, False, NoEventMask, &xevent);
				processedevent = 1;
				break;
			}
		}
	case MotionNotify:
		if ((event->xbutton.x < this->x) ||
				(event->xbutton.y < this->y) ||
				(event->xbutton.x > this->x+this->width) ||	
				(event->xbutton.y > this->y+this->height)) {
			if (this->pressed) {
				this->pressed = 0;
				processedevent = 1;
				Paint();
				break;
			}
		}
		break;
	case KeyPress:
		if(!this->focused)
			break;
		keysym = XKeycodeToKeysym(this->display, event->xkey.keycode, 0);
		strnzcpy(keystr, XKeysymToString(keysym), sizeof(keystr));
		// Only the space bar and the return key trigger some action:
		if ((strcmp(keystr, "space") == 0) ||(strcmp(keystr, "Return") == 0)) {
			this->pressed = 1;
			Paint();
			processedevent = 1;
		}
		break;
	case KeyRelease:
		if(!this->focused)
			break;
		keysym = XKeycodeToKeysym(this->display, event->xkey.keycode, 0);
		strnzcpy(keystr, XKeysymToString(keysym), sizeof(keystr));
		// Only the space bar and the return key trigger some action:
		if ((strcmp(keystr, "space") == 0) || (strcmp(keystr, "Return") == 0)) {
			this->pressed = 0;
			Paint();
			XSendEvent(this->display, this->window, False, NoEventMask, &xevent);
			processedevent = 1;
		}
		break;
	default:
		break;
	} // end switch(event.type)
	return(processedevent);
} // end int XButton::EventHandler(XEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XButton::Disabled()                                                  |
//|                                                                          |
//| Returns the status of the button's "disabled" field.                     |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XButton::Disabled() {
	return(this->disabled);
} // end of int XButton::Disabled()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XButton::Disable()                                                   |
//|                                                                          |
//| Sets this->disabled to 1. If this changes the button's state the button  |
//| is repainted.                                                            |
//| Returns 1.                                                               |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XButton::Disable() {
	if (this->disabled)
		return(1);
	this->disabled = 1;
	Paint();
	return(1);
} // end of int XButton::Disable()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XButton::Enable()                                                    |
//|                                                                          |
//| Sets this->disabled to 0. If this changes the button's state the button  |
//| is repainted.                                                            |
//| Returns 1.                                                               |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XButton::Enable() {
	if (!this->disabled)
		return(1);
	this->disabled = 0;
	Paint();
	return(1);
} // end of int XButton::Enable()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XStatusBar::XStatusBar()                                                 |
//|                                                                          |
//| The default constructor.                                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
XStatusBar::XStatusBar() {
} // end of XStatusBar::XStatusBar()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XStatusBar::XStatusBar(...) : XControl(), XCollection                    |
//|                                                                          |
//| Initialize the statusbar's fields and draw the (as yet blank) statusbar. |
//|                                                                          |
//+--------------------------------------------------------------------------+
XStatusBar::XStatusBar(XWinArgs* winargs, XProperties* properties) :
	XControl(winargs, properties), XCollection() {
	// Initialize members:
	this->type     = XW_TYPESTATUSBAR;
	this->y        = this->xwindow->Height() - this->height;
	this->width    = this->xwindow->Width();
	this->taborder = XW_NOTABRING;

	// Shrink the window's clipping region by the size of the statusbar:
	this->xwindow->ShrinkCanvas(XW_CANVASBOTTOM, this->height);

	// Add to the parent's collection of controls:
	if (!((XWindow*)this->parent)->Add(this)) {
		SetError("Statusbar could not be added to parent collection.");
		return;
	}

	// Draw the statusbar:
	this->Paint();
} // end of XStatusBar::XStatusBar(...) : XControl(), XCollection

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XStatusBar::Paint()                                                 |
//|                                                                          |
//| Paints the statusbar.                                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XStatusBar::Paint() {
	// Reset clipping region to max:
	this->xwindow->UnsetClipRegion();

	// In case the size of the parent window has changed:
	if ((this->xwindow->Width() != this->width) ||
			(this->xwindow->Height()-this->height != this->y)) {
		XSetFillStyle(this->display, this->gc, FillSolid);
		XSetForeground(this->display, this->gc, this->xwindow->BGColor());
		XFillRectangle(this->display, this->xwindow->videomem, this->gc,
									 this->x, this->y, this->width, this->height);
		XCopyArea(this->display, this->xwindow->videomem, this->window, this->gc,
							this->x, this->y, this->width, this->height, this->x, this->y);
		this->width = this->xwindow->Width();
		this->y     = this->xwindow->Height() - this->height;
		CalculateItemPositions();
	}

	// Set clipping region to parent window:
	this->xwindow->SetClipRegionToWindow();

	// Draw the statusbar background:
  XSetFillStyle(this->display, this->gc, FillSolid);
 	XSetForeground(this->display, this->gc, this->backgroundcolor);
	XFillRectangle(this->display, this->xwindow->videomem, this->gc,
								 this->x, this->y, this->width, this->height);

	// Draw the statusbar edge:
  XSetLineAttributes(this->display, this->gc, 1,
										 LineSolid, CapNotLast, JoinBevel);
	for (int i = 0; i < this->edgesize; ++i) {
		XSetForeground(this->display, this->gc, this->edgecolorlight);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x,             this->y+i,
							this->x+this->width-1, this->y+i);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+i, this->y,
							this->x+i, this->y+this->height-1);
		XSetForeground(this->display, this->gc, this->edgecolordark);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+i+1,           this->y+this->height-1-i,
							this->x+this->width-1, this->y+this->height-1-i);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+this->width-1-i, this->y+i+1,
							this->x+this->width-1-i, this->y+this->height-1);
	}


	// Copy the statusbar back onto main window:
	XCopyArea(this->display, this->xwindow->videomem, this->window, this->gc,
						this->x, this->y, this->width, this->height, this->x, this->y);

	// Finally draw the statusbar elements:
	this->collection->Reset();
	while(this->collection->Item() != NULL) {
		this->collection->Item()->Paint();
		this->collection->Next();
	}
} // end of void XStatusBar::Paint()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XStatusBar::Add(XControl* control)                                   |
//|                                                                          |
//| Adds a new item to the collection of statusbar items.                    |
//| Returns 1 if the item could be added and 0 in case of errors.            |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XStatusBar::Add(XControl* control) {
	// Reset error flag:
	ClearError();

	// Sanity checks:
	if (control == NULL) {
		SetError("Statusbar item is null pointer.");
		return(0);
	}
	if (control->Type() != XW_TYPESTATUSBARITEM) {
		SetError("Statusbar item is of wrong type.");
		return(0);
	}

	// Adding is done in the base class:
	if (!XCollection::Add(control))
		return(0);

	// Recalculate all variable width's and positions of collection:
	CalculateItemPositions();

	// Paint the new statusbar:
	Paint();
	return(1);
} // end of int XStatusBar::Add(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XStatusBar::Remove(XControl* control)                                |
//|                                                                          |
//| Removes an item from the collection of statusbar items.                  |
//| Returns 1 if the item could be removed and 0 in case of errors.          |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XStatusBar::Remove(XControl* control) {
	// Reset error flag:
	ClearError();

	// Sanity checks:
	if (control == NULL) {
		SetError("Statusbar item is null pointer.");
		return(0);
	}
	if (control->Type() != XW_TYPESTATUSBARITEM) {
		SetError("Statusbar item is of wrong type.");
		return(0);
	}

	// Removing is done in the base class:
	if (!XCollection::Remove(control))
		return(0);

	// Recalculate all variable width's and positions of collection:
	CalculateItemPositions();

	// Paint the new statusbar:
	Paint();
	return(1);
} // end of int XStatusBar::Remove(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XStatusBar::CalculateItemPositions()                                |
//|                                                                          |
//| Determines width of each item that has not set it, and the needed x's.   |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XStatusBar::CalculateItemPositions() {
	// Do we have any items?
	if (this->collection->Count() == 0)
		return;

	// Sum up the fixed width's:
	int totalfixedwidth = 0;
	int varitemscount = 0;
	this->collection->Reset();
	while (this->collection->Item() != NULL) {
		if (((XStatusBarItem*)this->collection->Item())->VarWidth())
			++varitemscount;
		else
			totalfixedwidth += ((XStatusBarItem*)this->collection->Item())->Width();
		this->collection->Next();
	}

	// Calculate variable width:
	int totalspacingwidth = (this->collection->Count()-1) * this->itemspacing;
	int netwidth = this->width - (2*this->padding);
	netwidth -= totalfixedwidth; 		// total width of fixed width items
	netwidth -= totalspacingwidth;	// total width of space between items
	int varwidth = 0;
	if (varitemscount > 0)
		varwidth = netwidth / varitemscount;

	// Set dimensions of items:
	int itemx = 0;
	int itemy = this->y + this->padding;
	int itemwidth = 0;
	int itemheight = this->height - (2 * this->padding);
	int currentx = this->padding;
	XStatusBarItem* current = NULL;
	this->collection->Reset();
	while (this->collection->Item() != NULL) {
		current = (XStatusBarItem*)this->collection->Item();
		itemwidth = current->Width();
		itemx = currentx;
		if (current->VarWidth()) {
			itemwidth = varwidth;
			currentx += varwidth + this->itemspacing;
		}
		else
			currentx += current->Width() + this->itemspacing;
		current->SetDimensions(itemx, itemy, itemwidth, itemheight);
		this->collection->Next();
	}
} // end of void XStatusBar::CalculateItemPositions()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XStatusBarItem::XStatusBarItem()                                         |
//|                                                                          |
//| The default constructor.                                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
XStatusBarItem::XStatusBarItem() {
} // end of XStatusBarItem::XStatusBarItem()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XStatusBarItem::XStatusBarItem(...) : XControl(...)                      |
//|                                                                          |
//| The statusbaritem's fields are initialized and the item is added to the  |
//| statusbar.                                                               |
//|                                                                          |
//+--------------------------------------------------------------------------+
XStatusBarItem::XStatusBarItem(XWinArgs* winargs, XProperties* properties) :
	XControl(winargs, properties) {
	// Sanity check:
	if (properties->parent == NULL) {
		SetError("Statusbar item has no parent!");
		return;
	}

	// Initialize fields:
	this->type = XW_TYPESTATUSBARITEM;
	this->varwidth = 0;
	if (this->width == 0)
		this->varwidth = 1;

	// Add to the statusbar's collection of controls:
	if (!((XStatusBar*)this->parent)->Add(this)) {
		SetError("Item could not be added to statusbar collection.");
		return;
	}
} // end of XStatusBarItem::XStatusBarItem(...) : XControl(...)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XStatusBarItem::Paint()                                             |
//|                                                                          |
//| Paints the statusbar item.                                               |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XStatusBarItem::Paint() {
	// Sanity Check:
	if ((this->width == 0) || (Error() != NULL))
		return;

	// Restrict clipping region to the window:
	this->xwindow->SetClipRegionToWindow();

	// Draw the background of the item:
  XSetFillStyle(this->display, this->gc, FillSolid);
 	XSetForeground(this->display, this->gc, this->backgroundcolor);
	XFillRectangle(this->display, this->xwindow->videomem, this->gc,
								 this->x, this->y, this->width-1, this->height-1);

	// Draw the edge of the item:
  XSetLineAttributes(this->display, this->gc, 1,
										 LineSolid, CapNotLast, JoinBevel);
	for (int i = 0; i < this->edgesize; ++i) {
		XSetForeground(this->display, this->gc, this->edgecolordark);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x,             this->y+i,
							this->x+this->width-1, this->y+i);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+i, this->y,
							this->x+i, this->y+this->height-1);
		XSetForeground(this->display, this->gc, this->edgecolorlight);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+i+1,           this->y+this->height-1-i,
							this->x+this->width-1, this->y+this->height-1-i);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+this->width-1-i, this->y+i+1,
							this->x+this->width-1-i, this->y+this->height-1);
	}

	// Draw the item's text:
	int textheight = this->font->ascent + this->font->descent;
	int usableheight = this->height - (2*this->edgesize);
	int usablewidth = this->width - (2*this->edgesize);
	usablewidth -= (2*XW_STATUSBARITEMPADDING);
	int tx = this->x + this->edgesize + XW_STATUSBARITEMPADDING;
	int ty = this->y + this->edgesize + this->font->ascent;
	ty += (usableheight - textheight)/2;
	XSetForeground(this->display, this->gc, this->foregroundcolor);
	XSetFont(this->display, this->gc, this->font->fid);
	XDrawString(this->display, this->xwindow->videomem, this->gc,
							tx, ty, this->text, strlen(this->text));

	// Finally copy the statusbaritem back onto main window:
	XCopyArea(this->display, this->xwindow->videomem, this->window, this->gc,
						this->x, this->y, this->width, this->height, this->x, this->y);
} // end of void XStatusBarItem::Paint()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XStatusBarItem::SetDimensions(int x, int y, int width, int height)  |
//|                                                                          |
//| Sets this->x, this-y, this->width and this->height.                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XStatusBarItem::SetDimensions(int x, int y, int width, int height) {
	this->x = x;
	this->y = y;
	this->width  = width;
	this->height = height;
} // end of void XStatusBarItem::SetDimensions(int x, int y, ...)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XStatusBarItem::VarWidth()                                           |
//|                                                                          |
//| Returns this->varwidth.                                                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XStatusBarItem::VarWidth() {
	return(this->varwidth);
} // end of int XStatusBarItem::VarWidth()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XStatusBarItem::Width()                                              |
//|                                                                          |
//| Returns this->width.                                                     |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XStatusBarItem::Width() {
	return(this->width);
} // end of int XStatusBarItem::Width()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XMenuObject::XMenuObject()                                               |
//|                                                                          |
//| The default constructor.                                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
XMenuObject::XMenuObject() {
} // end of XMenuObject::XMenuObject()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XMenuObject::XMenuObject(...) : XControl(), XCollection()                |
//|                                                                          |
//| Sets this->taborder to XW_NOTABRING, this->activemenu to NULL and        |
//| this->status to XW_HIDEMENUENTRIES.                                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
XMenuObject::XMenuObject(XWinArgs* winargs, XProperties* properties) :
	XControl(winargs, properties), XCollection() {
	// Initialize fields:
	this->type       = XW_TYPEMENUOBJECT;
	this->taborder   = XW_NOTABRING;
	this->status     = XW_HIDEMENUENTRIES;
	this->activemenu = NULL;
} // end of XMenuObject::XMenuObject(...) : XControl(...), XCollection()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMenuObject::Paint()                                                |
//|                                                                          |
//| Interface definition.                                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMenuObject::Paint() {
} // end of void XMenuObject::Paint()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenuObject::EventHandler(XEvent* event)                             |
//|                                                                          |
//| Interface definition.                                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenuObject::EventHandler(XEvent* event) {
	return(1);
} // end of int XMenuObject::EventHandler(XEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenuObject::Add(XControl* control)                                  |
//|                                                                          |
//| Interface definition.                                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenuObject::Add(XControl* control) {
	return(1);
} // end of int XMenuObject::Add(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenuObject::Remove(XControl* control)                               |
//|                                                                          |
//| Interface definition.                                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenuObject::Remove(XControl* control) {
	return(1);
} // end of int XMenuObject::Remove(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMenuObject::SetStatus(int isactive)                                |
//|                                                                          |
//| Sets this->status.                                                       |
//+--------------------------------------------------------------------------+
void XMenuObject::SetStatus(int status) {
	this->status = status;
} // end of void XMenuObject::SetStatus(int status)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenuObject::SetActiveMenuTo(XMenu* item)                            |
//|                                                                          |
//| If this->activemenu is not NULL, this->activemenu->isactive is set to 0, |
//| the menu entries of this->activemenu are hidden and this->activemenu is  |
//| repainted.                                                               |
//| Then this->activemenu is set to item, this->activemenu->isactive is set  |
//| to 1, this->activemenu is repainted and it's entries are shown.          |
//| Returns 1 if the activemenu could succesfully be set and 0 if there was  |
//| a type mismatch with item (only objects of type XW_TYPEMENU are allowed).|
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenuObject::SetActiveMenuTo(XMenu* item) {
	// Sanity check:
	if ((item != NULL) && (item->Type() != XW_TYPEMENU))
		return(0);

	// Reset previously active menu (if there is one):
	if (this->activemenu != NULL) {
		// Hide any possibly open submenu dropdown lists:
		this->collection->Reset();
		while(this->collection->Item() != NULL) {
			if (((XMenu*)this->collection->Item())->Visible())
				((XMenu*)this->collection->Item())->SetActiveMenuTo(NULL);
			this->collection->Next();
		}
		this->activemenu->SetStatus(XW_HIDEMENUENTRIES);
		this->activemenu->HideDropDownList();
		this->activemenu->Paint();
	}

	// Set new active menu (if one is specified):
	this->activemenu = item;
	if (this->activemenu != NULL) {
		this->activemenu->SetStatus(XW_SHOWMENUENTRIES);
		this->activemenu->Paint();
		this->activemenu->ShowDropDownList();
	}
	return(1);
} // end of int XMenuObject::SetActiveMenuTo(XMenu* item)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XMenu* XMenuObject::ActiveMenu()                                         |
//|                                                                          |
//| Returns this->activemenu.                                                |
//|                                                                          |
//+--------------------------------------------------------------------------+
XMenu* XMenuObject::ActiveMenu() {
	return(this->activemenu);
} // end of XMenu* XMenuObject::ActiveMenu()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMenuObject::CalculateMenuPositions()                               |
//|                                                                          |
//| Interface definition.                                                    |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMenuObject::CalculateMenuPositions() {
} // end of void XMenuObject::CalculateMenuPositions()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XMenuBar::XMenuBar()                                                     |
//|                                                                          |
//| The default constructor.                                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
XMenuBar::XMenuBar() {
} // end of XMenuBar::XMenuBar()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XMenuBar::XMenuBar(...) : XMenuObject(...)                               |
//|                                                                          |
//| Initialize the menubar's fields and draw the (as yet blank) menubar.     |
//|                                                                          |
//+--------------------------------------------------------------------------+
XMenuBar::XMenuBar(XWinArgs* winargs, XProperties* properties) :
	XMenuObject(winargs, properties) {
	// Initialize fields:
	this->type     = XW_TYPEMENUBAR;
	this->width    = this->xwindow->Width();
	this->height   = XW_MENUBARHEIGHT;
	this->edgesize = XW_MENUBAREDGESIZE;

	// Shrink the window's clipping region by the size of the menubar:
	this->xwindow->ShrinkCanvas(XW_CANVASTOP, this->height);

	// Add to the parent's collection of controls:
	if (!((XWindow*)this->parent)->Add(this)) {
		SetError("Menubar could not be added to parent collection.");
		return;
	}

	// Draw the menubar:
	this->Paint();
} // end of XMenuBar::XMenuBar(...) : XMenuObject(...)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMenuBar::Paint()                                                   |
//|                                                                          |
//| Paints the menubar.                                                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMenuBar::Paint() {
	// In case the size of the parent window has changed:
	if (this->xwindow->Width() != this->width) {
		this->width = this->xwindow->Width();
		CalculateMenuPositions();
	}

	// Restrict clipping region to window:
	this->xwindow->SetClipRegionToWindow();

	// Draw the menubar background:
  XSetFillStyle(this->display, this->gc, FillSolid);
 	XSetForeground(this->display, this->gc, this->backgroundcolor);
	XFillRectangle(this->display, this->xwindow->videomem, this->gc,
								 this->x, this->y, this->width, this->height);

	// Draw the menubar edge:
  XSetLineAttributes(this->display, this->gc, 1,
										 LineSolid, CapNotLast, JoinBevel);
	for (int i = 0; i < this->edgesize; ++i) {
		XSetForeground(this->display, this->gc, this->edgecolorlight);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x,             this->y+i,
							this->x+this->width-1, this->y+i);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+i, this->y,
							this->x+i, this->y+this->height-1);
		XSetForeground(this->display, this->gc, this->edgecolordark);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+i+1,           this->y+this->height-1-i,
							this->x+this->width-1, this->y+this->height-1-i);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+this->width-1-i, this->y+i+1,
							this->x+this->width-1-i, this->y+this->height-1);
	}

	// Copy the menubar back onto main window:
	XCopyArea(this->display, this->xwindow->videomem, this->window, this->gc,
						this->x, this->y, this->width, this->height, this->x, this->y);

	// Finally draw the menus:
	this->collection->Reset();
	while(this->collection->Item() != NULL) {
		this->collection->Item()->Paint();
		this->collection->Next();
	}
} // end of void XMenuBar::Paint()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenuBar::EventHandler(XEvent* event)                                |
//|                                                                          |
//| The event handler for the menubar just hands down any events it receives |
//| to the menus under it's control.                                         |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenuBar::EventHandler(XEvent* event) {
	// If the collection is empty, we can return right now:
	if (this->collection->Count() == 0)
		return(0);

	// Iterate through the list of menus and hand down the event:
	this->collection->Reset();
	while(this->collection->Item() != NULL) {
		if (((XMenu*)this->collection->Item())->EventHandler(event))
			return(1);
		this->collection->Next();
	}

	// Ok, let's see if the menubar can do something with the event:
	int processedevent = 0;
	switch (event->type) {
	case ButtonPress:
		if (event->xbutton.button == Button1) {
			if (this->activemenu != NULL)
				SetActiveMenuTo(NULL);
		}
		break;
	default:
		break;
	} // end switch(event.type)
	return(processedevent);
} // end of int XMenuBar::EventHandler(XEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenuBar::Add(XControl* control)                                     |
//|                                                                          |
//| Adds a new menu to the collection of menues in the menubar.              |
//| Returns 1 if the item could be added and 0 in case of errors.            |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenuBar::Add(XControl* control) {
	// Reset error flag:
	ClearError();

	// Sanity checks:
	if (control == NULL) {
		SetError("Menubar menu is null pointer.");
		return(0);
	}
	if (control->Type() != XW_TYPEMENU) {
		SetError("Menubar menu is of wrong type.");
		return(0);
	}

	// Adding is done in the base class:
	if (!XCollection::Add(control))
		return(0);

	// Paint the new menubar:
	CalculateMenuPositions();
	Paint();
	return(1);
} // end of int XMenuBar::Add(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenuBar::Remove(XControl* control)                                  |
//|                                                                          |
//| Removes a menu from the collection of menus in the menubar.              |
//| Returns 1 if the menu could be removed and 0 in case of errors.          |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenuBar::Remove(XControl* control) {
	// Reset error flag:
	ClearError();

	// Sanity checks:
	if (control == NULL) {
		SetError("Menubar menu is null pointer.");
		return(0);
	}
	if (control->Type() != XW_TYPEMENU) {
		SetError("Menubar menu is of wrong type.");
		return(0);
	}

	// Removing is done in the base class:
	if (!XCollection::Remove(control))
		return(0);

	// Paint the new menubar:
	CalculateMenuPositions();
	Paint();
	return(1);
} // end of int XMenuBar::Remove(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMenuBar::CalculateMenuPositions()                                  |
//|                                                                          |
//| Determines the x position and width of each menu.                        |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMenuBar::CalculateMenuPositions() {
	// Do we have any items?
	if (this->collection->Count() == 0)
		return;

	// Set dimensions of menus:
	int menux = 0;
	int menuy = this->y + this->edgesize;
	int menuwidth = 0;
	int menuheight = this->height - (2*this->edgesize);
	int currentx = 0;
	XMenu* current = NULL;
	this->collection->Reset();
	while (this->collection->Item() != NULL) {
		current = (XMenu*)this->collection->Item();
		menux = currentx;
		menuwidth = current->TextWidth() + (2 * current->Padding());
		menuwidth += (2 * current->Edgesize());
		if (menux + menuwidth > this->width)
			menuwidth = 0;
		current->SetDimensions(menux, menuy, menuwidth, menuheight);
		currentx += menuwidth;
		this->collection->Next();
	}
} // end of void XMenuBar::CalculateMenuPositions()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XMenu::XMenu(...) : XMenuObject(...)                                     |
//|                                                                          |
//| The menu's fields are initialized and the menu is added to the menubar.  |
//|                                                                          |
//+--------------------------------------------------------------------------+
XMenu::XMenu(XWinArgs* winargs, XProperties* properties, XMenuBar* menubar) :
	XMenuObject(winargs, properties) {
	// Sanity check:
	if (properties->parent == NULL) {
		SetError("Menu has no parent!");
		return;
	}
	if (menubar == NULL) {
		SetError("Menu has no menubar!");
		return;
	}


	// Initialize fields:
	this->type = XW_TYPEMENU;
	this->status = XW_HIDEMENUENTRIES;
	this->dropdown = new XRectangle();
	this->menubar = menubar;
	this->visible = 0;
	if (this->parent->Type() == XW_TYPEMENUBAR)
		this->visible = 1; // menus in menubar always visible

	// Set the atom needed to communicate with the parent:
	if ((wm_menu_clicked =
			 XInternAtom(this->display, "WM_MENU_CLICKED", False)) == None) {
		SetError("can't set WM_MENU_CLICKED atom.");
		return;
	}

	// Add to the parent's collection of controls:
	int result = 0;
	if (this->parent->Type() == XW_TYPEMENUBAR)
		result = ((XMenuBar*)this->parent)->Add(this);
	else if (this->parent->Type() == XW_TYPEMENU)
		result = ((XMenu*)this->parent)->Add(this);
	if (result == 0) {
		SetError("Menu could not be added to parent collection.");
		return;
	}
} // end of XMenu::XMenu(...) : XMenuObject(...)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XMenu::~XMenu()                                                          |
//|                                                                          |
//| Frees the rectangle with the dropdown position and size.                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
XMenu::~XMenu() {
	XFree(this->dropdown);
} // end of XMenu::~XMenu()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMenu::Paint()                                                      |
//|                                                                          |
//| Draws the menu as entry in the menubar or in a menu (we have to write    |
//| both into the window and into the videomemory because we map the window  |
//| after this is drawn for the first time).                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMenu::Paint() {
	// Sanity Check:
	if ((this->width == 0) || (Error() != NULL))
		return;

	// Draw the background of the menu:
  XSetFillStyle(this->display, this->gc, FillSolid);
 	XSetForeground(this->display, this->gc, this->backgroundcolor);
	if (this->parent->Type() == XW_TYPEMENUBAR) {
		XFillRectangle(this->display, this->xwindow->videomem, this->gc,
									 this->x, this->y, this->width, this->height);
	}
	XFillRectangle(this->display, this->window, this->gc,
								 this->x, this->y, this->width, this->height);

	// If activated draw an edge around the menu:
	if (this->status == XW_SHOWMENUENTRIES) {
		XSetLineAttributes(this->display, this->gc, 1,
											 LineSolid, CapNotLast, JoinBevel);
		for (int i = 0; i < this->edgesize; ++i) {
			XSetForeground(this->display, this->gc, this->edgecolorlight);
			if (this->parent->Type() == XW_TYPEMENUBAR) {
				XDrawLine(this->display, this->xwindow->videomem, this->gc,
									this->x,             this->y+i,
									this->x+this->width-1, this->y+i);
				XDrawLine(this->display, this->xwindow->videomem, this->gc,
									this->x+i, this->y,
									this->x+i, this->y+this->height-1);
			}
			XDrawLine(this->display, this->window, this->gc,
								this->x,             this->y+i,
								this->x+this->width-1, this->y+i);
			XDrawLine(this->display, this->window, this->gc,
								this->x+i, this->y,
								this->x+i, this->y+this->height-1);
			XSetForeground(this->display, this->gc, this->edgecolordark);
			if (this->parent->Type() == XW_TYPEMENUBAR) {
				XDrawLine(this->display, this->xwindow->videomem, this->gc,
									this->x+i+1,           this->y+this->height-1-i,
									this->x+this->width-1, this->y+this->height-1-i);
				XDrawLine(this->display, this->xwindow->videomem, this->gc,
									this->x+this->width-1-i, this->y+i+1,
									this->x+this->width-1-i, this->y+this->height-1);
			}
			XDrawLine(this->display, this->window, this->gc,
								this->x+i+1,           this->y+this->height-1-i,
								this->x+this->width-1, this->y+this->height-1-i);
			XDrawLine(this->display, this->window, this->gc,
								this->x+this->width-1-i, this->y+i+1,
								this->x+this->width-1-i, this->y+this->height-1);
		}
	}

	// Draw the menu's text:
	int usableheight = this->height - (2*this->edgesize);
	int tx = this->x + this->edgesize + this->padding;
	int ty = this->y + this->edgesize + (usableheight/2);
	ty += this->font->ascent/2;
	XSetForeground(this->display, this->gc, this->foregroundcolor);
	XSetFont(this->display, this->gc, this->font->fid);
	if (this->parent->Type() == XW_TYPEMENUBAR) {
		XDrawString(this->display, this->xwindow->videomem, this->gc,
								tx, ty, this->text, strlen(this->text));
	}
	XDrawString(this->display, this->window, this->gc,
							tx, ty, this->text, strlen(this->text));

	// If this menu has submenu entries, draw a small triangle:
	if ((this->parent->Type() == XW_TYPEMENU) &&
			(this->collection->Count() > 0)) {
		XSetForeground(this->display, this->gc, this->edgecolorlight);
		XDrawLine(this->display, this->window, this->gc,
							this->x+this->width-this->padding,
							this->y+(this->height/2),
							this->x+this->width-this->padding-XW_SUBMENUINDICATORSIZE,
							this->y+(this->height/2)-(XW_SUBMENUINDICATORSIZE/2));
		XDrawLine(this->display, this->window, this->gc,
							this->x+this->width-this->padding-XW_SUBMENUINDICATORSIZE,
							this->y+(this->height/2)-(XW_SUBMENUINDICATORSIZE/2),
							this->x+this->width-this->padding-XW_SUBMENUINDICATORSIZE,
							this->y+(this->height/2)+(XW_SUBMENUINDICATORSIZE/2));
		XSetForeground(this->display, this->gc, this->edgecolordark);
		XDrawLine(this->display, this->window, this->gc,
							this->x+this->width-this->padding-XW_SUBMENUINDICATORSIZE,
							this->y+(this->height/2)+(XW_SUBMENUINDICATORSIZE/2),
							this->x+this->width-this->padding-1,
							this->y+(this->height/2));
	}
} // end of void XMenu::Paint()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenu::EventHandler(XEvent* event)                                   |
//|                                                                          |
//| The event handler for menu entries.                                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenu::EventHandler(XEvent* event) {
	// If the menu is not visible, we don't look for events:
	if (!this->visible)
		return(0);

	// Iterate through the list of menu entries and hand down the event:
	this->collection->Reset();
	while(this->collection->Item() != NULL) {
		if (((XMenu*)this->collection->Item())->EventHandler(event))
			return(1);
		this->collection->Next();
	}

	// Ok, let's see if the event was meant for this menu entry:
	int processedevent = 0;
	char keystr[XW_TEXTSIZESYSTEMMSG];
	memset(&keystr, 0, sizeof(keystr));
	KeySym keysym;
	XEvent xevent;
	memset(&xevent, 0, sizeof(xevent));
	xevent.xclient.type = ClientMessage;
	xevent.xclient.window = this->window;
	xevent.xclient.message_type = this->wm_menu_clicked;
	xevent.xclient.format = 32;
	strcpy(xevent.xclient.data.b, this->uuid);
	xevent.xclient.data.l[0] = this->wm_menu_clicked;
	xevent.xclient.data.l[1] = (long)this;
	switch (event->type) {
	case ButtonPress:
		// Only left mouse button does anything:
		if (event->xbutton.button != Button1)
			break;

		// Check if we are "on" a menu entry:
		if ((event->xbutton.x < this->x) ||
				(event->xbutton.y < this->y) ||
				(event->xbutton.x > this->x+this->width) ||	
				(event->xbutton.y > this->y+this->height))
			break;

		// Set this menu as active if this is not the active one:
		if (this->status != XW_SHOWMENUENTRIES) {
			((XMenuObject*)this->parent)->SetActiveMenuTo(this);
			processedevent = 1;
			break;
		}

		// This is the active one so hide all submenu dropdowns:
		if (this->activemenu != NULL) {
			SetActiveMenuTo(NULL);
			processedevent = 1;
			break;
		}
		processedevent = 1;
		break;
	case ButtonRelease:
		// Only left mouse button does anything:
		if (event->xbutton.button != Button1)
			break;

		// Button up on an entry triggers the "menu clicked" event:
		if ((event->xbutton.x > this->x) &&
				(event->xbutton.x < this->x+this->width) &&
				(event->xbutton.y > this->y) &&
				(event->xbutton.y < this->y+this->height)) {
			if ((this->collection->Count() == 0) &&
					(this->status == XW_SHOWMENUENTRIES)) {
				XSendEvent(this->display, this->window, False, NoEventMask, &xevent);
				this->menubar->SetActiveMenuTo(NULL);
				processedevent = 1;
			}
			else {
				processedevent = 1;
			}
			break;
		}
		break;
	case MotionNotify:
		// Only left mouse button does anything:
		if (event->xbutton.state != Button1Mask)
			break;

		// If motion switches to adjacent menu depends on parent:
		if (this->parent->Type() == XW_TYPEMENUBAR) {
			// In the menubar, only sideways motion switches menus:
			if ((event->xbutton.x > this->x) &&
					(event->xbutton.x < this->x+this->width)) {
				if ((event->xbutton.y > this->y) &&
						(event->xbutton.y < this->y+this->height)) {
					if (this->status != XW_SHOWMENUENTRIES) {
						((XMenuObject*)this->parent)->SetActiveMenuTo(this);
						processedevent = 1;
						break;
					}
				}
			}
		}
		else {
			// In a normal menu only vertical motion switches menus:
			if ((event->xbutton.y < this->y) ||
					(event->xbutton.y > this->y+this->height)) {
				if ((event->xbutton.x > this->x) &&
						(event->xbutton.x < this->x+this->width)) {
					if (this->status == XW_SHOWMENUENTRIES) {
						((XMenuObject*)this->parent)->SetActiveMenuTo(NULL);
						processedevent = 1;
						break;
					}
				}
			}
			else {
				if ((event->xbutton.x > this->x) &&
						(event->xbutton.x < this->x+this->width)) {
					if (this->status != XW_SHOWMENUENTRIES) {
						((XMenuObject*)this->parent)->SetActiveMenuTo(this);
						processedevent = 1;
						break;
					}
				}
			}
		} // end if (this->parent->Type() == XW_TYPEMENUBAR) ... else

		// Horizontal motion outside dropdowns unsets an active submenu:
		if (this->status == XW_SHOWMENUENTRIES) {
			if ((this->activemenu != NULL) &&
					(this->activemenu->ChildCount() == 0)) {
				if ((event->xbutton.x < this->dropdown->x) ||
						(event->xbutton.x > this->dropdown->x + this->dropdown->width)) {
					SetActiveMenuTo(NULL);
					processedevent = 1;
					break;
				}
			}
		}
		break;
	case KeyPress:
		if (this->status != XW_SHOWMENUENTRIES)
			break;
		keysym = XKeycodeToKeysym(this->display, event->xkey.keycode, 0);
		strnzcpy(keystr, XKeysymToString(keysym), sizeof(keystr));
		if (strcmp(keystr, "Escape") == 0) {
			if (this->parent->Type() == XW_TYPEMENUBAR)
				((XMenuObject*)this->parent)->SetActiveMenuTo(NULL);
			else
				((XMenuObject*)((XControl*)this->parent)->Parent())->SetActiveMenuTo(NULL);				
			processedevent = 1;
		}
		break;
	case KeyRelease:
		break;
	default:
		break;
	} // end switch(event.type)
	return(processedevent);
} // end int XMenu::EventHandler(XEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenu::Add(XControl* control)                                        |
//|                                                                          |
//| Adds an entry to the collection of entries in the menu.                  |
//| Returns 1 if the entry could be added and 0 in case of errors.           |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenu::Add(XControl* control) {
	// Reset error flag:
	ClearError();

	// Sanity checks:
	if (control == NULL) {
		SetError("Menu entry is null pointer.");
		return(0);
	}
	if ((control->Type() != XW_TYPEMENU) &&
			(control->Type() != XW_TYPEMENUSEPARATOR))	{
		SetError("Menu entry is of wrong type.");
		return(0);
	}

	// Adding is done in the base class:
	if (!XCollection::Add(control))
		return(0);

	return(1);
} // end of int XMenu::Add(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenu::Remove(XControl* control)                                     |
//|                                                                          |
//| Removes an entry from the collection of entries in a menu.               |
//| Returns 1 if the entry could be removed and 0 in case of errors.         |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenu::Remove(XControl* control) {
	// Reset error flag:
	ClearError();

	// Sanity checks:
	if (control == NULL) {
		SetError("Menu entry is null pointer.");
		return(0);
	}
	if (control->Type() != XW_TYPEMENU) {
		SetError("Menu entry is of wrong type.");
		return(0);
	}

	// Removing is done in the base class:
	if (!XCollection::Remove(control))
		return(0);

	return(1);
} // end of int XMenu::Remove(XControl* control)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMenu::SetDimensions(int x, int y, int width, int height)           |
//|                                                                          |
//| Sets this->x, this-y, this->width and this->height.                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMenu::SetDimensions(int x, int y, int width, int height) {
	this->x = x;
	this->y = y;
	this->width  = width;
	this->height = height;
} // end of void XMenu::SetDimensions(int x, int y, ...)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMenu::SetVisibility(int visibility)                                |
//|                                                                          |
//| Sets this->visible to visibility.                                        |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMenu::SetVisibility(int visibility) {
	this->visible = visibility;
} // end of void XMenu::SetVisibility(int visibility)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenu::ShowDropDownList()                                            |
//|                                                                          |
//| Draws the drop down list with the entries.                               |
//| Returns 1.                                                               |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenu::ShowDropDownList() {
	// If there are no entries, we don't have to do anything:
	if (this->collection->Count() == 0)
		return(1);

	// Determine the width of the drop down list:
	int maxwidth = 0;
	int curwidth = 0;
	int ptrwidth = 0;
	XMenu* current = NULL;
	this->collection->Reset();
	while (this->collection->Item() != NULL) {
		current = (XMenu*)this->collection->Item();
		curwidth = current->TextWidth() + (2 * current->Edgesize());
		curwidth += (2 * current->Padding());
		if (curwidth > maxwidth)
			maxwidth = curwidth;
		if (current->ChildCount() > 0)
			ptrwidth = XW_SUBMENUINDICATORSIZE + (2 * current->Padding());
		this->collection->Next();
	}

	// Allow for the submenu indicator if needed:
	maxwidth += ptrwidth;

	// Set the dimensions of the entries in the list:
	this->dropdown->x = this->x;
	if (this->parent->Type() == XW_TYPEMENU)
		this->dropdown->x += this->width + this->edgesize;
	this->dropdown->y = this->y;
	if (this->parent->Type() == XW_TYPEMENUBAR)
		this->dropdown->y += this->height-1;
	this->dropdown->width  = maxwidth;
	this->dropdown->height = 0;
	int menux = this->dropdown->x + this->edgesize;
	int menuy = 0;
	int menuwidth  = maxwidth;
	int menuheight = 0;
	int currenty = this->dropdown->y + this->edgesize;
	this->collection->Reset();
	while (this->collection->Item() != NULL) {
		current = (XMenu*)this->collection->Item();
		menuy = currenty;
		if (current->Type() == XW_TYPEMENUSEPARATOR)
			menuheight = 2;
		else {
			current->SetVisibility(1);
			menuheight = XW_MENUVSPACE + (2 * current->Edgesize());
			menuheight += current->TextHeight();
			if ((ptrwidth > 0) &&
					(XW_SUBMENUINDICATORSIZE > current->TextHeight()))
				menuheight += (XW_SUBMENUINDICATORSIZE - current->TextHeight());
		}
		current->SetDimensions(menux, menuy, menuwidth, menuheight);
		this->dropdown->height += menuheight;
		currenty += menuheight;
		this->collection->Next();
	}

	// Allow for the edge:
	this->dropdown->width += (2 * this->edgesize);
	this->dropdown->height += (2 * this->edgesize);

	// Reset clipping region to full window area:
	this->xwindow->SetClipRegionToWindow();

	// Draw the background of the drop down list:
  XSetFillStyle(this->display, this->gc, FillSolid);
 	XSetForeground(this->display, this->gc, this->backgroundcolor);
	XFillRectangle(this->display, this->window, this->gc,
								 this->dropdown->x, this->dropdown->y,
								 this->dropdown->width, this->dropdown->height);

	// If activated draw an edge around the menu:
	if (this->status == XW_SHOWMENUENTRIES) {
		XSetLineAttributes(this->display, this->gc, 1,
											 LineSolid, CapNotLast, JoinBevel);
		for (int i = 0; i < this->edgesize; ++i) {
			XSetForeground(this->display, this->gc, this->edgecolorlight);
			XDrawLine(this->display, this->window, this->gc,
								this->dropdown->x,
								this->dropdown->y+i,
								this->dropdown->x+this->dropdown->width-1,
								this->dropdown->y+i);
			XDrawLine(this->display, this->window, this->gc,
								this->dropdown->x+i,
								this->dropdown->y,
								this->dropdown->x+i,
								this->dropdown->y+this->dropdown->height-1);
			XSetForeground(this->display, this->gc, this->edgecolordark);
			XDrawLine(this->display, this->window, this->gc,
								this->dropdown->x+i+1,
								this->dropdown->y+this->dropdown->height-1-i,
								this->dropdown->x+this->dropdown->width-1,
								this->dropdown->y+this->dropdown->height-1-i);
			XDrawLine(this->display, this->window, this->gc,
								this->dropdown->x+this->dropdown->width-1-i,
								this->dropdown->y+i+1,
								this->dropdown->x+this->dropdown->width-1-i,
								this->dropdown->y+this->dropdown->height-1);
		}
	}

	// Draw the entries:
	this->collection->Reset();
	while(this->collection->Item() != NULL) {
		this->collection->Item()->Paint();
		this->collection->Next();
	}
	return(1);
} // end of int XMenu::ShowDropDownList()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenu::HideDropDownList()                                            |
//|                                                                          |
//| Restores the state of the application before the drop down list was      |
//| opened, i.e. draws what was under the drop down list.                    |
//| Returns 1.                                                               |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenu::HideDropDownList() {
	// Reset visibility of entries in dropdown list:
	this->collection->Reset();
	while (this->collection->Item() != NULL) {
		if (this->collection->Item()->Type() == XW_TYPEMENU)
			((XMenu*)this->collection->Item())->SetVisibility(0);
		this->collection->Next();
	}

	// Restore background:
	XCopyArea(this->display, this->xwindow->videomem, this->window, this->gc,
						this->dropdown->x, this->dropdown->y,
						this->dropdown->width, this->dropdown->height,
						this->dropdown->x, this->dropdown->y);
	return(1);
} // end of int XMenu::HideDropDownList()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenu::Visible()                                                     |
//|                                                                          |
//| Returns this->visible.                                                   |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenu::Visible() {
	return(this->visible);
} // end of int XMenu::Visible()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XMenuSeparator::XMenuSeparator(...) : XMenuObject(...)                   |
//|                                                                          |
//| The separator's fields are initialized and the separator is added to the |
//| menuparent collection.                                                   |
//|                                                                          |
//+--------------------------------------------------------------------------+
XMenuSeparator::XMenuSeparator(XWinArgs* winargs,
															 XProperties* properties,
															 XMenuBar* menubar) :
	XMenuObject(winargs, properties) {
	// Sanity check:
	if (properties->parent == NULL) {
		SetError("Separator has no parent!");
		return;
	}
	if (properties->parent->Type() == XW_TYPEMENUBAR) {
		SetError("Separator can't be added to menubar!");
		return;
	}
	if (menubar == NULL) {
		SetError("Separator has no menubar!");
		return;
	}


	// Initialize fields:
	this->type = XW_TYPEMENUSEPARATOR;
	this->menubar = menubar;

	// Add to the parent's collection of controls:
	if(!((XMenu*)this->parent)->Add(this)) {
		SetError("Separator could not be added to parent collection.");
		return;
	}
} // end of XMenuSeparator::XMenuSeparator(...) : XMenuObject(...)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMenuSeparator::Paint()                                             |
//|                                                                          |
//| Draws the engraved line.                                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMenuSeparator::Paint() {
	// Sanity Check:
	if ((this->width == 0) || (Error() != NULL))
		return;

	// Draw the engraved line:
	XSetForeground(this->display, this->gc, this->edgecolordark);
	XDrawLine(this->display, this->window, this->gc,
						this->x,               this->y,
						this->x+this->width-1, this->y);
	XSetForeground(this->display, this->gc, this->edgecolorlight);
	XDrawLine(this->display, this->window, this->gc,
						this->x,               this->y+1,
						this->x+this->width-1, this->y+1);
} // end of void XMenuSeparator::Paint()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XMenuSeparator::EventHandler(XEvent* event)                          |
//|                                                                          |
//| The event handler for the separator.                                     |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XMenuSeparator::EventHandler(XEvent* event) {
	int processedevent = 0;
	switch (event->type) {
	case ButtonPress:
		// Only left mouse button does anything:
		if (event->xbutton.button != Button1)
			break;

		// Button down on a separator -> do nothing but eliminate event:
		if ((event->xbutton.x >= this->x) &&
				(event->xbutton.x < this->x+this->width) &&
				(event->xbutton.y >= this->y) &&
				(event->xbutton.y < this->y+this->height)) {
			processedevent = 1;
			break;
		}
		break;
	case ButtonRelease:
		// Only left mouse button does anything:
		if (event->xbutton.button != Button1)
			break;

		// Button up on a separator -> close all dropdown lists:
		if ((event->xbutton.x >= this->x) &&
				(event->xbutton.x < this->x+this->width) &&
				(event->xbutton.y >= this->y) &&
				(event->xbutton.y < this->y+this->height)) {
			this->menubar->SetActiveMenuTo(NULL);
			processedevent = 1;
			break;
		}
		break;
	default:
		break;
	} // end switch(event.type)
	return(processedevent);
} // end int XMenuSeparator::EventHandler(XEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XMenuSeparator::SetDimensions(int x, int y, int width, int height)  |
//|                                                                          |
//| Sets this->x, this-y, this->width and this->height.                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XMenuSeparator::SetDimensions(int x, int y, int width, int height) {
	this->x = x;
	this->y = y;
	this->width  = width;
} // end of void XMenuSeparator::SetDimensions(int x, int y, ...)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XTextBox::XTextBox()                                                     |
//|                                                                          |
//| The default constructor.                                                 |
//|                                                                          |
//+--------------------------------------------------------------------------+
XTextBox::XTextBox() {
} // end of XTextBox::XTextBox()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| XTextBox::XTextBox(...) : XControl(...)                                  |
//|                                                                          |
//| The textbox's fields are initialized and the textbox is added to the     |
//| window.                                                                  |
//|                                                                          |
//+--------------------------------------------------------------------------+
XTextBox::XTextBox(XWinArgs* winargs, XProperties* properties) :
	XControl(winargs, properties) {
	// Sanity check:
	if (properties->parent == NULL) {
		SetError("Textbox has no parent!");
		return;
	}

	// Initialize fields:
	this->type = XW_TYPETEXTBOX;
	this->cursorpos = 0;
	this->isactive = 0;
	this->lastsec = time(NULL);
	this->currsec = this->lastsec;
	this->textx = this->x + this->edgesize + this->padding;
	this->texty = this->y + this->font->ascent;
	this->texty += (this->height - (this->font->ascent+this->font->descent))/2;

	// Set a new background color:
	XColor rgb;
	XColor xcol;
	XAllocNamedColor(this->display, this->colormap, "Gainsboro", &rgb, &xcol);
 	this->backgroundcolor = rgb.pixel;

	// Set the atom needed to communicate with the parent:
	if ((wm_text_changed =
			 XInternAtom(this->display, "WM_TEXT_CHANGED", True)) == None) {
		SetError("Can't set WM_TEXT_CHANGED atom.");
		return;
	}

	// Add to the window's collection of controls:
	if (!((XWindow*)this->parent)->Add(this)) {
		SetError("Textbox could not be added to window collection.");
		return;
	}

	// Draw the textbox:
	this->Paint();
} // end of XTextBox::XTextBox(...) : XControl(...)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XTextBox::Paint()                                                   |
//|                                                                          |
//| Paints the textbox.                                                      |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XTextBox::Paint() {
	// Sanity Check:
	if (Error() != NULL)
		return;

	// Restrict clipping region to the window:
	//this->xwindow->SetClipRegionToWindow();

	// Draw the background of the textbox:
  XSetFillStyle(this->display, this->gc, FillSolid);
 	XSetForeground(this->display, this->gc, this->backgroundcolor);
	XFillRectangle(this->display, this->xwindow->videomem, this->gc,
								 this->x, this->y, this->width-1, this->height-1);

	// Draw the edge of the textbox:
  XSetLineAttributes(this->display, this->gc, 1,
										 LineSolid, CapNotLast, JoinBevel);
	for (int i = 0; i < this->edgesize; ++i) {
		XSetForeground(this->display, this->gc, this->edgecolordark);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x,             this->y+i,
							this->x+this->width-1, this->y+i);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+i, this->y,
							this->x+i, this->y+this->height-1);
		XSetForeground(this->display, this->gc, this->edgecolorlight);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+i+1,           this->y+this->height-1-i,
							this->x+this->width-1, this->y+this->height-1-i);
		XDrawLine(this->display, this->xwindow->videomem, this->gc,
							this->x+this->width-1-i, this->y+i+1,
							this->x+this->width-1-i, this->y+this->height-1);
	}

	// Finally copy the textbox back onto main window:
	XCopyArea(this->display, this->xwindow->videomem, this->window, this->gc,
						this->x, this->y, this->width, this->height, this->x, this->y);
} // end of void XTextBox::Paint()

//+--------------------------------------------------------------------------+
//|                                                                          |
//| int XTextBox::EventHandler(XEvent* event)                                |
//|                                                                          |
//| The event handler for the textbox.                                       |
//|                                                                          |
//+--------------------------------------------------------------------------+
int XTextBox::EventHandler(XEvent* event) {
	int processedevent = 0;

	// The button is not disabled -> process events:
	char keystr[XW_TEXTSIZESYSTEMMSG];
	memset(&keystr, 0, sizeof(keystr));
	KeySym keysym;
	XEvent xevent;
	memset(&xevent, 0, sizeof(xevent));
	xevent.xclient.type = ClientMessage;
	xevent.xclient.window = this->window;
	xevent.xclient.message_type = this->wm_text_changed;
	xevent.xclient.format = 32;
	strcpy(xevent.xclient.data.b, this->uuid);
	xevent.xclient.data.l[0] = this->wm_text_changed;
	xevent.xclient.data.l[1] = (long)this;
	unsigned int i;
	switch (event->type) {
	case ButtonPress:
		if (event->xbutton.button == Button1) {
			if ((event->xbutton.x > this->x) &&
					(event->xbutton.y > this->y) &&
					(event->xbutton.x < this->x+this->width) &&
					(event->xbutton.y < this->y+this->height)) {
				this->isactive = 1;
				DrawCursor(1);
			}
			else {
				this->isactive = 0;
				if (this->cursorvisible)
					DrawCursor(0);
			}
			processedevent = 1;
			break;
		}
		break;
	case ButtonRelease:
		break;
	case MotionNotify:
		break;
	case KeyPress:
		if(!this->isactive)
			break;
		keysym = XKeycodeToKeysym(this->display, event->xkey.keycode, 0);
		strnzcpy(keystr, XKeysymToString(keysym), sizeof(keystr));
		fprintf(stderr, "Key pressed \"%s\"\n", keystr);
		if (strcmp(keystr, "Escape") == 0) {
			this->isactive = 0;
			if (this->cursorvisible)
				DrawCursor(0);
			processedevent = 1;
			break;
		}
		if (strcmp(keystr, "Left") == 0) { // key: [<-]
			if (this->cursorpos > 0) {
				DrawCursor(0);
				--this->cursorpos;
				DrawCursor(1);
			}
			processedevent = 1;
			break;
		}
		if (strcmp(keystr, "Right") == 0) { // key: [->]
			if (this->cursorpos < strlen(this->text)) {
				DrawCursor(0);
				++this->cursorpos;
				DrawCursor(1);
			}
			processedevent = 1;
			break;
		}
		if (strcmp(keystr, "Home") == 0) { // key: [Home]
			if (this->cursorpos > 0) {
				DrawCursor(0);
				this->cursorpos = 0;
				DrawCursor(1);
			}
			processedevent = 1;
			break;
		}
		if (strcmp(keystr, "End") == 0) {
			if (this->cursorpos < strlen(this->text)) {
				DrawCursor(0);
				this->cursorpos = strlen(this->text);
				DrawCursor(1);
			}
			processedevent = 1;
			break;
		}
		if (strcmp(keystr, "Delete") == 0) {
			if (this->cursorpos == strlen(this->text)) {
				processedevent = 1;
				break;
			}
			DrawCursor(0);
			for (i = cursorpos; i < strlen(this->text); ++i)
				this->text[i] = this->text[i+1];
			DrawCursor(1);
			processedevent = 1;
			break;
		}
		if (strcmp(keystr, "BackSpace") == 0) {
			if (this->cursorpos == 0) {
				processedevent = 1;
				break;
			}
			DrawCursor(0);
			for (i = cursorpos-1; i < strlen(this->text); ++i)
				this->text[i] = this->text[i+1];
			--this->cursorpos;
			DrawCursor(1);
			processedevent = 1;
			break;
		}
		break;
	case KeyRelease:
		break;
	default:
		break;
	} // end switch(event.type)

	// If this textbox is active, let the cursor blink:
	if (this->isactive) {
		this->currsec = time(NULL);
		if (this->currsec > this->lastsec) {
			this->lastsec = this->currsec;
			if (this->cursorvisible)
				DrawCursor(0);
			else
				DrawCursor(1);
		}
	}
	return(processedevent);
} // end int XTextBox::EventHandler(XEvent* event)

//+--------------------------------------------------------------------------+
//|                                                                          |
//| void XTextBox::DrawCursor(int visible)                                   |
//|                                                                          |
//| Draws the cursor at index this->cursorpos in the color determinded by    |
//| the value passed in show: if that value is one, the forground color is   |
//| used and the cursor made visible, if the value is other than 1 the back- |
//| ground color is used, making the cursor invisible.                       |
//|                                                                          |
//+--------------------------------------------------------------------------+
void XTextBox::DrawCursor(int visible) {
	// Clear the background of the textbox:
  XSetFillStyle(this->display, this->gc, FillSolid);
 	XSetForeground(this->display, this->gc, this->backgroundcolor);
	XFillRectangle(this->display, this->xwindow->videomem, this->gc,
								 this->x+this->edgesize, this->y+this->edgesize,
								 this->width-1-(2*this->edgesize),
								 this->height-1-(2*this->edgesize));

	// Draw the item's text again:
	XSetForeground(this->display, this->gc, this->foregroundcolor);
	XSetFont(this->display, this->gc, this->font->fid);
	XDrawString(this->display, this->xwindow->videomem, this->gc,
							this->textx, this->texty, this->text, strlen(this->text));

	// Determine which color to use for the cursor:
	if (visible) {
		this->cursorvisible = 1;
		XSetForeground(this->display, this->gc, this->foregroundcolor);
	}
	else {
		this->cursorvisible = 0;
		XSetForeground(this->display, this->gc, this->backgroundcolor);
	}
	
	// Draw the cursor:
	int cursorx = this->textx;
	if (this->cursorpos > 0) {
		char buf[XW_TEXTSIZECONTROL];
		memset(buf, 0, sizeof(buf));
		if (this->cursorpos < strlen(this->text)) {
			for(unsigned int i = 0; i < this->cursorpos; ++i)
				buf[i] = this->text[i];
		}
		else {
			strnzcpy(buf, this->text, sizeof(buf));
		}
		cursorx += XTextWidth(this->font, buf, strlen(buf));
	}	
	XDrawLine(this->display, this->xwindow->videomem, this->gc,
						cursorx, this->texty-this->font->ascent,
						cursorx, this->texty+this->font->descent);
	XDrawLine(this->display, this->window, this->gc,
						cursorx, this->texty-this->font->ascent,
						cursorx, this->texty+this->font->descent);

	// And copy the textbox back onto main window agin:
	XCopyArea(this->display, this->xwindow->videomem, this->window, this->gc,
						this->x, this->y, this->width, this->height, this->x, this->y);

	return;
} // end of void XTextBox::DrawCursor(int show)
