//+--------------------------------------------------------------------------
//|
//| xwindow.h - declaration of various x window related classes
//|
//| This file is part of a larger package; see the README supplied with
//| this for information about the package, how to build and how to run
//| it.
//|
//| (c) 2002, 2020 Dietmar Koenig <http://www.faoileag.de>
//|
//| This program is free software: you can redistribute it and/or modify
//| it under the terms of the GNU General Public License as published by
//| the Free Software Foundation, either version 3 of the License, or
//| (at your option) any later version.
//|
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//| GNU General Public License for more details.
//|
//| You should have received a copy of the GNU General Public License
//| along with this program.  If not, see <https://www.gnu.org/licenses/>.
//|
//+--------------------------------------------------------------------------
#include <X11/Intrinsic.h>
#include <X11/Xlocale.h>
#include <iostream>

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Defines needed by the various classes:                                   |
//|                                                                          |
//+--------------------------------------------------------------------------+
#define XW_MENUBARHEIGHT             23
#define XW_MENUBAREDGESIZE            1
#define XW_MENUBARPADDING             5
#define XW_MENUEDGESIZE               2
#define XW_MENUPADDING                5
#define XW_MENUVSPACE                 2
#define XW_SUBMENUINDICATORSIZE       8
#define XW_FRAMESIZEFOCUSED           2
#define XW_FRAMESIZENORMAL            1
#define XW_STATUSBARHEIGHT           26
#define XW_STATUSBAREDGESIZE          1
#define XW_STATUSBARPADDING           3
#define XW_STATUSBARITEMSPACING       4
#define XW_STATUSBARITEMEDGESIZE      1
#define XW_STATUSBARITEMPADDING       5
#define XW_TEXTBOXPADDING             5
#define XW_MAXWIDTH                1600
#define XW_MAXHEIGHT               1200
#define XW_MINWIDTH                  32
#define XW_MINHEIGHT                 32

#define XW_NOTABRING                 -1
#define XW_CANVASLEFT                 1
#define XW_CANVASRIGHT                2
#define XW_CANVASTOP                  3
#define XW_CANVASBOTTOM               4

#define XW_TEXTSIZECONTROL         1024
#define XW_TEXTSIZEERRORMSG        1024
#define XW_TEXTSIZEFONTINFO         512
#define XW_TEXTSIZESYSTEMMSG        256
#define XW_TEXTSIZETYPE             256
#define XW_TEXTSIZENAME             256
#define XW_TEXTSIZEUUID              17

#define XW_TYPEOBJECT                 1
#define XW_TYPEWINDOW                 2
#define XW_TYPEAPPLICATION            3
#define XW_TYPECONTROL                4
#define XW_TYPETEXTCONTROL            5
#define XW_TYPEBUTTON                30
#define XW_TYPESTATUSBAR             20
#define XW_TYPESTATUSBARITEM         21
#define XW_TYPEMENUOBJECT            10
#define XW_TYPEMENUBAR               11
#define XW_TYPEMENU                  12
#define XW_TYPEMENUSEPARATOR         13
#define XW_TYPETEXTBOX               40

#define XW_SHOWMENUENTRIES            1
#define XW_HIDEMENUENTRIES            0

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Declare classes:                                                         |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XList;					// a linked list of XObjects
class XLoop;					// a looped linked list of XObjects
class XObject;				// a base class for all windows and controls
class XWindow;				// base for windows
class XAppRoot;				// base for applications (together with XWindow)
class XControl;				// base for all controls
class XButton;				// a button control
class XStatusBar;			// a statusbar control
class XStatusBarItem;	// an item in the statusbar
class XMenuObject;    // base object for XMenuBar and XMenu
class XMenuBar;       // a menu bar
class XMenu;          // menus for XMenuBar

//+--------------------------------------------------------------------------+
//|                                                                          |
//| This structure forms the base of XList and XLoop:                        |
//|                                                                          |
//+--------------------------------------------------------------------------+
typedef struct TList {
	XControl* control;
	TList* next;
	TList* prev;
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XList                                  A linked list of XControls. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XList {
public:
	XList();														// initializes all pointers to NULL
	virtual ~XList();										// frees the memory used by the list
	virtual int Add(XControl* item);		// adds an item to the list
	virtual int Insert(XControl* item);	// insert an item before the cursor
	virtual int Remove(XControl* item);	// removes an item from the list
	virtual int Clear();                // removes all items from the list
	virtual int Reset();								// resets the cursor to the first item
	virtual int Next();									// moves the cursor to the next item
	virtual int Previous();							// moves the cursor to the previous item
	virtual XControl* Item();  					// returns pointer to item under cursor
	virtual int Count();								// returns number of items in the list

protected:
	int count;
	TList* cursor;
	TList* first;
	TList* last;
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XTabRing         A sorted loop of XControls (soerted by taborder). |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XTabRing : public XList {
public:
	XTabRing();															// initializes all pointers to NULL
	virtual int Insert(XControl* item);			// sorts item into the tab ring
	virtual int Remove(XControl* item);			// removes item from the tab ring
	virtual int SetFocusTo(XControl* item);	// sets focus and cursor to item
	virtual int Clear();										// removes all controls from tab ring
	virtual int Reset();										// resets focus / cursor to first
	virtual int Next();											// moves cursor to next control
	virtual int Previous();									// moves cursor to previous control
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XWinArgs                A class with the basic x window arguments. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XWinArgs {
 public:
	XWinArgs();
	Display* display;
	Window window;
	GC gc;	
	Colormap colormap;
	int screen;
	XWindow* xwindow;
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XProperties   A class with all of a control's possible properties. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XProperties {
public:
	XProperties();
	int x;
	int y;
	int width;
	int height;
	int edgesize;
	int padding;
	int taborder;
	int itemspacing;
	char* text;
	XFontStruct* font;
	XObject* parent;
	void Reset();
	int SizeOfText();

protected:
	char textbuf[XW_TEXTSIZECONTROL];
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XObject                         The base class for all my objects. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XObject {
public:
	XObject();
	XObject(XWinArgs* winargs);
	virtual ~XObject();
	int Type();
	char* Error();
	char* Uuid();
	void ClearError();
	void SetError(char* errormsg);

protected:
	void Initialize();
	char error[XW_TEXTSIZEERRORMSG];
	char uuid[XW_TEXTSIZEUUID];
	int type;
	int x, y;
	int width, height;
	int screen;
	Display* display;
	Window window;
	GC gc;
	Colormap colormap;
	XRectangle* canvas;
	XWindow* xwindow;
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XCollection             Base class for objects with child objects. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XCollection {
public:
	XCollection();
	virtual ~XCollection();
	virtual int Add(XControl* control);
	virtual int Remove(XControl* control);
	int ChildCount();

protected:
	XList* collection; // a collection of child objects
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XWindow                                    The basic window class. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XWindow : public XObject, public XCollection {
	public:
	XWindow();
	XWindow(int x, int y,
					int width,
					int height,
					char* title,
					int argc, char** argv);
	virtual ~XWindow();
	virtual int Run();
	virtual int Add(XControl* control);
	virtual int Remove(XControl* control);
	void SetClipRegionToCanvas();
	void SetClipRegionToWindow();
	void UnsetClipRegion();
	int SetFocusTo(XControl* control);
	int ShrinkCanvas(int side, int shrink);
	int ExpandCanvas(int side, int expand);
	int Width();
	int Height();
	unsigned long BGColor();
	Pixmap videomem;
	XIC ic;
	XFontSet fontset;

  protected:
	virtual void Paint();
	virtual void Terminate();
	virtual void EventLoopAction();
	virtual void EvHandlerClientMessage(XClientMessageEvent* event);
	virtual void EvHandlerButton1Press(XButtonEvent* event);
	virtual void EvHandlerButton2Press(XButtonEvent* event);
	virtual void EvHandlerButton3Press(XButtonEvent* event);
	virtual void EvHandlerButton1Release(XButtonEvent* event);
	virtual void EvHandlerButton2Release(XButtonEvent* event);
	virtual void EvHandlerButton3Release(XButtonEvent* event);
	virtual void EvHandlerButton1MotionNotify(XButtonEvent* event);
	virtual void EvHandlerButton2MotionNotify(XButtonEvent* event);
	virtual void EvHandlerButton3MotionNotify(XButtonEvent* event);
	virtual void EvHandlerKeyPressed(XKeyEvent* event);
	virtual void EvHandlerKeyReleased(XKeyEvent* event);
	XIMStyle BestMatchedStyle(XIMStyle style1, XIMStyle style2);

	unsigned long bgcolor;		// color of the window's background
	int terminate;						// if 1 the event loop is no longer executed
	int eventloopaction;			// if 1 EventLoopAction() is called from event loop
	int shiftdown;						// if 1 <Shift> is currently held down
	int altdown;							// if 1 <Alt> is currently held down
	int altgrdown;						// if 1 <Alt Gr> is currently held down
	int ctrldown;							// if 1 <Strg> is currently held down
	Atom wm_delete_window;		// "close" message from window manager
	Atom wm_button_clicked;		// "clicked" message from a button
	Atom wm_menu_clicked;			// "clicked" message from a menu entry
	Atom wm_text_changed;			// if the text in a textbox is changed
	XTabRing* tabring;				// collection of controls focusable by tab
	XSizeHints* sizehints;
	XWMHints* wmhints;
	XClassHint* classhints;
	XTextProperty windowname;
};  // end of class XWindow

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XAppRoot           The base class for the basic application class. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XAppRoot {
public:
	XAppRoot();
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XApplication                  The base class for all applications. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XApplication : public XAppRoot, public XWindow {
public:
	XApplication();
	XApplication(int x, int y,
							 int width,
							 int height,
							 char* title,
							 int argc, char** argv);
	virtual int Run();
protected:
	char progname[XW_TEXTSIZENAME];
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XControl         The base class for all my controls (buttons etc). |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XControl : public XObject {
public:
	XControl();
	XControl(XWinArgs* winargs, XProperties* properties);
	virtual ~XControl();
	virtual void Paint();
	virtual int EventHandler(XEvent* event);
	XObject* Parent();
	char* Text();
	void SetText(char* text);
	void SetFocused();
	void SetNormal();
	int TextWidth();
	int TextHeight();
	int TabOrder();
	int Focused();
	int Padding();
	int Edgesize();

protected:
	virtual void AllocateColors();	// allocates system colors (background etc).
	void SetClipRegionToObject();		// sets clipping region to outer rectangle
	void SetClipRegionToCanvas();		// sets clipping region to inner rectangle
	unsigned long framecolor;				// color of frame around control
	unsigned long edgecolorlight;		// color of spotlight side of edges
	unsigned long edgecolordark;		// color of edges in shadow
	unsigned long backgroundcolor;	// color of the control's background
	unsigned long foregroundcolor;	// color of the control's text
	unsigned long disabledcolor;		// color of text when control is disabled
	int framesize;									// normally 1; bolder if focused
	int focused;										// 1 if control has focus, 0 otherwise
	int edgesize;                   // size of the edge of a control
	int padding;                    // padding of text
	int taborder;                   // position of control in the tab ring
	int itemspacing;                // horizontal space between items
	char text[XW_TEXTSIZECONTROL];  // the text a control might have
	XFontStruct* font;							// font information for above text
	XObject* parent;								// parent object of a control
};


//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XButton                           A class for rectangular buttons. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XButton : public XControl {
public:
	XButton();
	XButton(XWinArgs* winargs, XProperties* properties);
	virtual void Paint();
	virtual int EventHandler(XEvent* event);
	int Disabled();
	int Disable();
	int Enable();

protected:
	Atom wm_button_clicked; // for communication with parent
	int disabled;						// 1 if disabled and 0 otherwise
	int pressed;						// 1 if pressed and 0 otherwise
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XStatusBar                                    The statusbar class. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XStatusBar : public XControl, public XCollection {
public:
	XStatusBar();
	XStatusBar(XWinArgs* winargs, XProperties* properties);
	virtual void Paint();
	virtual int Add(XControl* control);
	virtual int Remove(XControl* control);

protected:
	void CalculateItemPositions();
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XStatusBarItem                  The items making up the statusbar. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XStatusBarItem : public XControl {
public:
	XStatusBarItem();
	XStatusBarItem(XWinArgs* winargs, XProperties* properties);
	virtual void Paint();
	void SetDimensions(int x, int y, int width, int height);
	int VarWidth();
	int Width();

protected:
	int varwidth; // 1 if item is of variable size and 0 if not
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XMenuObject                 The base class for XMenuBar and XMenu. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XMenuObject : public XControl, public XCollection {
public:
	XMenuObject();
	XMenuObject(XWinArgs* winargs, XProperties* properties);
	virtual void Paint();
	virtual int EventHandler(XEvent* event);
	virtual int Add(XControl* control);
	virtual int Remove(XControl* control);
	void SetStatus(int status);
	int SetActiveMenuTo(XMenu* item);
	XMenu* ActiveMenu();

protected:
	virtual void CalculateMenuPositions();
	XMenu* activemenu;
	int status;
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XMenuBar             XMenuBar is a container class for menu items. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XMenuBar : public XMenuObject {
public:
	XMenuBar();
	XMenuBar(XWinArgs* winargs, XProperties* properties);
	virtual void Paint();
	virtual int EventHandler(XEvent* event);
	virtual int Add(XControl* control);
	virtual int Remove(XControl* control);

protected:
	virtual void CalculateMenuPositions();
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XMenu                                      The menus for XMenuBar. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XMenu : public XMenuObject {
public:
	XMenu(XWinArgs* winargs, XProperties* properties, XMenuBar* menubar);
	virtual ~XMenu();
	virtual void Paint();
	virtual int EventHandler(XEvent* event);
	virtual int Add(XControl* control);
	virtual int Remove(XControl* control);
	void SetDimensions(int x, int y, int width, int height);
	void SetVisibility(int visibility);
	int ShowDropDownList();
	int HideDropDownList();
	int Visible();

protected:
	Atom wm_menu_clicked;			// "clicked" message from a menu entry
	XRectangle* dropdown;			// this rectangle defines the drop down area
	XMenuBar* menubar;				// the menubar this menu belongs to
	int visible;							// 1 if this menu is visible, 0 if hidden
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XMenuSeparator                              A separator for menus. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XMenuSeparator : public XMenuObject {
public:
	XMenuSeparator(XWinArgs* winargs,
								 XProperties* properties,
								 XMenuBar* menubar);
	virtual void Paint();
	virtual int EventHandler(XEvent* event);
	void SetDimensions(int x, int y, int width, int height);

protected:
	XMenuBar* menubar;				// the menubar this menu belongs to
};

//+--------------------------------------------------------------------------+
//|                                                                          |
//| Class XTextBox                                  A textbox to enter text. |
//|                                                                          |
//+--------------------------------------------------------------------------+
class XTextBox : public XControl {
public:
	XTextBox();
	XTextBox(XWinArgs* winargs, XProperties* properties);
	virtual void Paint();
	virtual int EventHandler(XEvent* event);

protected:
	void DrawCursor(int visible);
	Atom wm_text_changed; // for communication with parent
	int textx;						// x value of begin of text
	int texty;						// y value of baseline of text
	int isactive;					// 1 if this textbox is set to receive input
	int cursorvisible;		// visibility of cursor: 1 cursor is visible
	unsigned int cursorpos;	// position of cursor in characters from first
	time_t lastsec;				// to determine when cursor blinks (every second)
	time_t currsec;
};
